"""Test the RedbackWrapperModel."""

import numpy as np
import pytest
from citation_compass import find_in_citations
from lightcurvelynx.math_nodes.given_sampler import GivenValueList
from lightcurvelynx.models.redback_models import RedbackWrapperModel


class ToySNModel:
    """A toy model that mimics the data structure of a RedbackTimeSeriesSource,
    but we can control.

    Attributes
    ----------
    self.height : float
        The height of the toy model peak.
    self.width : float
        The width of the toy model peak.
    """

    def __init__(self, height, width):
        self.height = height
        self.width = width

    def minwave(self):
        """Get the minimum wavelength of the model."""
        return 0.0

    def maxwave(self):
        """Get the maximum wavelength of the model."""
        return np.inf

    def get_flux_density(self, times, wavelengths):
        """A toy flux function that depends on time and wave.
        Peaks at t=0 and decreases with wave.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps.
        wavelengths : numpy.ndarray
            A length N array of wavelengths (in angstroms).

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        lightcurve = self.height * np.exp(-(times**2) / (self.width**2))
        flux_density = lightcurve[:, np.newaxis] * (1000.0 / wavelengths[np.newaxis, :])
        return flux_density


def _toy_redback_model(times, height, width, **kwargs):
    """Create and return the toy model."""
    return ToySNModel(height, width)


# Fake the appending of a citation to the model function.
_toy_redback_model.citation = "TEST_CITATION_2025"


def test_redback_models_toy() -> None:
    """Test that we can create and evaluate a simple model."""
    # Define static parameters.
    t0 = 64350.0
    parameters = {
        "height": 1000.0,
        "width": 10.0,
    }

    # Create the model.
    model = RedbackWrapperModel(
        _toy_redback_model,
        parameters=parameters,  # Set ALL the redback model parameters
        ra=0.0,  # Set other parameters
        dec=-10.0,
        t0=t0,
        node_label="source",
    )
    assert set(model.source_param_names) == {"height", "width"}

    state = model.sample_parameters()
    assert state["source"]["height"] == 1000.0
    assert state["source"]["width"] == 10.0
    assert state["source"]["t0"] == 64350.0

    times = np.array([-10.0, 0.5, 10.0]) + t0
    waves_ang = np.array([1000.0, 2000.0])
    fluxes = model.evaluate_sed(times, waves_ang, graph_state=state)

    # Check that the fluxes spike around t0.
    assert fluxes.shape == (3, 2)
    assert np.all(fluxes[0, :] < fluxes[1, :])
    assert np.all(fluxes[1, :] > fluxes[2, :])

    # Check that the fluxes are different at different wavelengths.
    assert np.all(fluxes[:, 0] != fluxes[:, 1])

    # Check that we can recover the citations
    rb_citations = find_in_citations("RedbackWrapperModel")
    assert len(rb_citations) >= 1
    for citation in rb_citations:
        assert "https://ui.adsabs.harvard.edu/abs/2024MNRAS.531.1203S/abstract" in citation
    rb_model_citations = find_in_citations("redback model")
    assert len(rb_model_citations) >= 1
    for citation in rb_model_citations:
        assert "TEST_CITATION_2025" in citation


def test_redback_models_fail_toy() -> None:
    """Test that we can create, but fail to evaluate a model if we don't have the parameters we need."""
    # Create a model with a missing parameter.
    t0 = 64350.0
    model = RedbackWrapperModel(
        _toy_redback_model,
        parameters={"height": 1000.0},  # Missing "width"
        t0=t0,
        node_label="source",
    )
    assert set(model.source_param_names) == {"height"}

    state = model.sample_parameters()
    times = np.array([1.0, 5.0, 6.5, 10.0])
    waves_ang = np.array([1000.0, 2000.0])
    with pytest.raises(TypeError):
        _ = model.evaluate_sed(times, waves_ang, graph_state=state)

    # Fail if we give it the same parameter in two different ways (repeat redshift).
    with pytest.raises(ValueError):
        _ = RedbackWrapperModel(
            "one_component_kilonova_model",
            parameters={"height": 1000.0, "width": 10.0, "redshift": 0.05},
            redshift=0.1,
            node_label="toy",
        )


def test_redback_models_chained_toy() -> None:
    """Test that we can create and evaluate a model with chained parameters."""
    t0 = 64350.0
    parameters = {
        "height": GivenValueList([500.0, 1000.0, 1500.0]),
        "width": 10.0,
    }

    # Create the model.
    model = RedbackWrapperModel(
        _toy_redback_model,
        parameters=parameters,  # Set ALL the redback model parameters
        t0=t0,
        node_label="source",
    )
    assert set(model.source_param_names) == {"height", "width"}

    state = model.sample_parameters(num_samples=3)
    times = np.array([t0 + 0.5])
    waves_ang = np.array([1000.0])
    fluxes = model.evaluate_sed(times, waves_ang, graph_state=state)

    # The returned fluxes should all be different since height is changing.
    assert fluxes.shape == (3, 1, 1)
    assert len(np.unique(fluxes)) == 3



import numpy as np
from lightcurvelynx.astro_utils.noise_model import poisson_bandflux_std
from numpy.testing import assert_allclose


def test_poisson_flux_std_flux():
    """Test poisson_flux_std for photon noise dominated regime"""
    # The results should be independent of some input parameters
    rng = np.random.default_rng(None)

    flux = 10 ** rng.uniform(0.0, 5.0, 100)
    expected_flux_err = np.sqrt(flux)

    flux_err = poisson_bandflux_std(
        bandflux=flux,
        total_exposure_time=rng.uniform(),
        exposure_count=rng.integers(1, 100),
        psf_footprint=rng.uniform(),
        sky=0.0,
        zp=1.0,
        readout_noise=0.0,
        dark_current=0.0,
    )

    assert_allclose(flux_err, expected_flux_err, rtol=1e-10)


def test_poisson_flux_std_sky():
    """Test poisson_flux_std for sky noise dominated regime"""
    # The results should be independent of some input parameters
    rng = np.random.default_rng(None)

    n = 100

    zp = 10.0
    sky = 10 ** rng.uniform(-2.0, 2.0, n)
    psf_footprint = 10 ** rng.uniform(0.0, 2.0, n)
    expected_flux_err = np.sqrt(sky * psf_footprint) * zp

    flux_err = poisson_bandflux_std(
        bandflux=0.0,
        total_exposure_time=rng.uniform(),
        exposure_count=rng.integers(1, 100),
        psf_footprint=psf_footprint,
        sky=sky,
        zp=zp,
        readout_noise=0.0,
        dark_current=0.0,
    )

    assert_allclose(flux_err, expected_flux_err, rtol=1e-10)


def test_poisson_flux_std_readout():
    """Test poisson_flux_std for readout noise dominated regime"""
    # The results should be independent of some input parameters
    rng = np.random.default_rng(None)

    n = 100

    readout_noise = 10 ** rng.uniform(-2.0, 2.0, n)
    psf_footprint = 10 ** rng.uniform(0.0, 2.0, n)
    exposure_count = rng.integers(1, 100, n)
    expected_flux_err = readout_noise * np.sqrt(psf_footprint) * np.sqrt(exposure_count)

    flux_err = poisson_bandflux_std(
        bandflux=0.0,
        total_exposure_time=rng.uniform(),
        exposure_count=exposure_count,
        psf_footprint=psf_footprint,
        sky=0.0,
        zp=1.0,
        readout_noise=readout_noise,
        dark_current=0.0,
    )

    assert_allclose(flux_err, expected_flux_err, rtol=1e-10)


def test_poisson_flux_std_dark():
    """Test poisson_flux_std for dark current noise dominated regime"""
    # The results should be independent of some input parameters
    rng = np.random.default_rng(None)

    n = 100

    dark_current = 10 ** rng.uniform(-2.0, 2.0, n)
    total_exposure_time = rng.uniform(1.0, 3.0, n)
    psf_footprint = 10 ** rng.uniform(0.0, 2.0, n)

    dark_current_total = dark_current * total_exposure_time * psf_footprint
    expected_flux_err = np.sqrt(dark_current_total)

    flux_err = poisson_bandflux_std(
        bandflux=0.0,
        total_exposure_time=total_exposure_time,
        exposure_count=rng.integers(1, 100, n),
        psf_footprint=psf_footprint,
        sky=0.0,
        zp=1.0,
        readout_noise=0.0,
        dark_current=dark_current,
    )

    assert_allclose(flux_err, expected_flux_err, rtol=1e-10)


def readout_noise_function(exptime):
    """Readout noise function"""
    return 10.0 * exptime


def test_readout_noise_from_function():
    """Test poisson_flux_std for readout noise dominated regime,
    with readout noise given as a function of exposure times"""
    # The results should be independent of some input parameters
    rng = np.random.default_rng(None)

    n = 100

    psf_footprint = 10 ** rng.uniform(0.0, 2.0, n)
    exposure_time = rng.uniform(30, 100, n)
    expected_flux_err = readout_noise_function(exptime=exposure_time)

    flux_err = poisson_bandflux_std(
        bandflux=0.0,
        total_exposure_time=exposure_time,
        exposure_count=1,
        psf_footprint=psf_footprint,
        sky=0.0,
        zp=1.0,
        readout_noise=readout_noise_function,
        dark_current=0.0,
    )

    assert_allclose(flux_err, expected_flux_err, rtol=1e-10)



"""A wrapper around pzflow sampling.

For the full pzflow package see:
https://github.com/jfcrenshaw/pzflow
"""

import numpy as np
import pandas as pd
from citation_compass import CiteClass

from lightcurvelynx.base_models import FunctionNode


class PZFlowNode(FunctionNode, CiteClass):
    """A node that wraps sampling from pzflow.

    References
    ----------
    * Paper: Crenshaw et. al. 2024 - https://ui.adsabs.harvard.edu/abs/2024AJ....168...80C
    * Zenodo: Crenshaw et. al. 2024 - https://doi.org/10.5281/zenodo.10710271

    Attributes
    ----------
    flow : pzflow.flow.Flow or pzflow.flowEnsemble.FlowEnsemble
        The object from which to sample.
    columns : list of str
        The column names for the output columns.
    conditional_cols : list of str
        The names of the conditional columns used when sampling the flow.
    """

    def __init__(self, flow_obj, node_label=None, **kwargs):
        self.flow = flow_obj

        # Add each of the flow's data columns as an output parameter.
        self.columns = [x for x in flow_obj.data_columns]

        # Add each of the flow's conditional columns as an input parameter and
        # check that the column is include in the kwargs (and thus has an input).
        if flow_obj.conditional_columns is not None and len(flow_obj.conditional_columns) > 0:
            self.conditional_cols = []
            for col in flow_obj.conditional_columns:
                if col not in kwargs:
                    raise ValueError(f"Missing input parameter '{col}' needed by flow model.")
                self.conditional_cols.append(col)
        else:
            self.conditional_cols = None

        super().__init__(self._non_func, node_label=node_label, outputs=self.columns, **kwargs)

    @classmethod
    def from_file(cls, filename, node_label=None, **kwargs):
        """Create a PZFlowNode from a saved flow file.

        Parameters
        ----------
        filename : str or Path
            The location of the saved flow.
        node_label : str
            An optional human readable identifier (name) for the current node.
        **kwargs : dict, optional
            Additional function arguments, including the input parameters for the flow.
        """
        try:
            from pzflow import Flow
        except ImportError as err:
            raise ImportError(
                "pzflow package is not installed be default. You can install it with "
                "`pip install pzflow` or `conda install conda-forge::pzflow`."
            ) from err

        flow_to_use = Flow(file=filename)
        return PZFlowNode(flow_to_use, node_label=node_label, **kwargs)

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Return the given values.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            Unused in this function, but included to provide consistency with other
            compute functions.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        results : any
            The result of the computation. This return value is provided so that testing
            functions can easily access the results.
        """
        # If a random number generator is used, use that to compute the seed.
        seed = None if rng_info is None else int.from_bytes(rng_info.bytes(4), byteorder="big")

        # If the flow has conditional columns, extract them from the GraphState and use
        # them to sample the flow (with num_samples=1).
        if self.conditional_cols is not None and len(self.conditional_cols) > 0:
            all_params = self.get_local_params(graph_state)
            input_params = {}
            for col in self.conditional_cols:
                input_params[col] = all_params[col]
            input_df = pd.DataFrame(input_params)

            samples = self.flow.sample(1, conditions=input_df, seed=seed)
        else:
            samples = self.flow.sample(graph_state.num_samples, seed=seed)

        # Parse out each output column in the flow samples as its own result vector.
        results = []
        for attr_name in self.flow.data_columns:
            attr_values = samples[attr_name].values
            if graph_state.num_samples == 1:
                results.append(attr_values[0])
            else:
                results.append(np.array(attr_values))

        # Save and return the results. If we only have a single output column,
        # return that directly.
        if len(self.flow.data_columns) == 1:
            results = results[0]
        self._save_results(results, graph_state)
        return results



import numpy as np

from lightcurvelynx.astro_utils.noise_model import poisson_bandflux_std
from lightcurvelynx.consts import GAUSS_EFF_AREA2FWHM_SQ
from lightcurvelynx.obstable.obs_table import ObsTable


class FakeObsTable(ObsTable):
    """A subclass for a (simplified) fake survey. The user must provide a constant
    flux error to use or enough information to compute the poisson_bandflux_std noise model.
    To compute the flux error, the user must provide the following values
    either in the table or as keyword arguments to the constructor 1) sky, 2) zp_per_band,
    and 3) either a) fwhm_px or b) psf_footprint.

    Defaults are set for other parameters (e.g. exptime, nexposure, read_noise, dark_current), which
    the user can override with keyword arguments to the constructor.

    Parameters
    ----------
    table : dict or pandas.core.frame.DataFrame
        The table with all the ObsTable information.  Must have columns
        "time", "ra", "dec", and "filter".
    colmap : dict, optional
        A mapping of standard column names to their names in the input table.
    zp_per_band : dict
        A dictionary mapping filter names to their instrumental zero points (flux in nJy
        corresponding to 1 electron per exposure). The filters provided must match those
        in the table. This is required if the table does not have a zero point column.
    const_flux_error : float or dict, optional
        If provided, use this constant flux error (in nJy) for all observations (overriding
        the normal noise compuation). If a dictionary is provided, it should map filter names
        to constant flux errors per-band. This should only be used for testing purposes.
    dark_current : float, optional
        The dark current for the camera in electrons per second per pixel (default=0.0).
    exptime : float, optional
        The exposure time for the camera in seconds, used for dark current calculation only
        (default=30).
    psf_footprint : float, optional
        The effective psf_footprint of the PSF in pixels.
    fwhm_px : float or dict, optional
        The full-width at half-maximum of the PSF in pixels. If a dictionary is provided,
        it should map filter names to fwhm values. This is only needed if `psf_footprint` is not provided
        and `const_flux_error` is not provided (default=None).
    nexposure : int, optional
        The number of exposures per observation (default=1).
    radius : float, optional
        The angular radius of the field of view of the observations in degrees (default=None).
    read_noise : float, optional
        The read noise for the camera in electrons (default=0.0).
    sky : float or dict, optional
        The sky background in the units of electrons / pixel^2. If a dictionary is provided,
        it should map filter names to sky values. This is only needed if `const_flux_error`
        is not provided (default=None).
    **kwargs : dict
        Additional keyword arguments to pass to the ObsTable constructor. This includes overrides
        for survey parameters such as:
        - survey_name: The name of the survey (default="FAKE_SURVEY").
    """

    # Default survey values.
    _default_survey_values = {
        "dark_current": 0,
        "exptime": 30,  # seconds
        "fwhm_px": None,  # pixels
        "nexposure": 1,  # exposures
        "radius": None,  # degrees
        "read_noise": 0,  # electrons
        "sky": None,  # electrons / pixel^2
        "survey_name": "FAKE_SURVEY",
    }

    def __init__(
        self,
        table,
        *,
        colmap=None,
        zp_per_band=None,
        const_flux_error=None,
        dark_current=0,
        exptime=30,
        psf_footprint=None,
        fwhm_px=None,
        nexposure=1,
        radius=None,
        read_noise=0,
        sky=None,
        **kwargs,
    ):
        self.zp_per_band = zp_per_band
        self.const_flux_error = const_flux_error

        # Pass along all the survey parameters to the parent class.
        super().__init__(
            table,
            colmap=colmap,
            dark_current=dark_current,
            exptime=exptime,
            psf_footprint=psf_footprint,
            fwhm_px=fwhm_px,
            nexposure=nexposure,
            radius=radius,
            read_noise=read_noise,
            sky=sky,
            **kwargs,
        )

        if const_flux_error is not None:
            # Convert a constant into a per-band dictionary.
            if isinstance(const_flux_error, int | float):
                self.const_flux_error = {fil: const_flux_error for fil in self.filters}

            # Check that every filter occurs in the dictionary with a non-negative value.
            for fil in self.filters:
                if fil not in self.const_flux_error:
                    raise ValueError(
                        "`const_flux_error` must include all the filters in the table. Missing '{fil}'."
                    )
            for fil, val in self.const_flux_error.items():
                if val < 0:
                    raise ValueError(f"Constant flux error for band {fil} must be non-negative. Got {val}.")
        else:
            # Make sure we have the required columns (fwhm_px, sky, exptime, nexposure) to
            # compute the flux error. If any are missing, assign a constant column from the survey values.
            if not self._assign_constant_if_needed("exptime", check_positive=True):
                raise ValueError("Must provide `exptime` to FakeSurveyTable.")
            if not self._assign_constant_if_needed("nexposure", check_positive=True):
                raise ValueError("Must provide `nexposure` to FakeSurveyTable.")
            if not self._assign_constant_if_needed("psf_footprint", check_positive=True):
                if not self._assign_constant_if_needed("fwhm_px", check_positive=True):
                    raise ValueError("Must provide either `psf_footprint` or `fwhm_px` to FakeSurveyTable.")
                psf_footprint = GAUSS_EFF_AREA2FWHM_SQ * (self._table["fwhm_px"]) ** 2
                self.add_column("psf_footprint", psf_footprint)
            if not self._assign_constant_if_needed("sky", check_positive=False):
                raise ValueError("Must provide `sky` to FakeSurveyTable.")

    def _assign_zero_points(self):
        """Assign instrumental zero points in nJy to the ObsTable. In this fake
        survey, we use a constant zero point per band.
        """
        # Check that we either have previously assigned zero points or have been given
        # a dictionary of zero points per band.
        if "zp" in self._table.columns:
            return  # Already have a column of zero points.
        if self.zp_per_band is None:
            raise ValueError("Must provide `zp_per_band` to FakeSurveyTable without a column of zero points.")
        if "filter" not in self._table.columns:
            raise ValueError(
                "Must provide a `filter` column to FakeSurveyTable without a column of zero points."
            )

        # Check that we have a zero point for every filter in the table.
        for fil in self.filters:
            if fil not in self.zp_per_band:
                raise ValueError(f"Must provide a zero point for filter {fil} in `zp_per_band`.")

        # Create a column of zero points, setting all values based on the filter column.
        zp_col = np.zeros(len(self._table), dtype=float)
        for key, val in self.zp_per_band.items():
            if val <= 0:
                raise ValueError(f"Zero point for band {key} must be positive. Got {val}.")
            zp_col[self._table["filter"] == key] = val
        self.add_column("zp", zp_col, overwrite=True)

    def bandflux_error_point_source(self, bandflux, index):
        """Compute observational bandflux error for a point source

        Parameters
        ----------
        bandflux : array_like of float
            Band bandflux of the point source in nJy.
        index : array_like of int
            The index of the observation in the OpSim table.

        Returns
        -------
        flux_err : array_like of float
            Simulated bandflux noise in nJy.
        """
        # If we have a constant flux error, use that.
        if self.const_flux_error is not None:
            filters = self._table["filter"].iloc[index]
            return np.array([self.const_flux_error[fil] for fil in filters])

        # Otherwise compute the flux error using the poisson_bandflux_std noise model.
        # We insert most the needed columns during construction, so we
        # can look up most of the values needed for the noise model.
        observations = self._table.iloc[index]
        psf_footprint = observations["psf_footprint"]

        return poisson_bandflux_std(
            bandflux,
            total_exposure_time=observations["exptime"],
            exposure_count=observations["nexposure"],
            psf_footprint=psf_footprint,
            sky=observations["sky"],
            zp=observations["zp"],
            readout_noise=self.safe_get_survey_value("read_noise"),
            dark_current=self.safe_get_survey_value("dark_current"),
        )



import numpy as np
import pytest
from lightcurvelynx.astro_utils.passbands import PassbandGroup
from lightcurvelynx.graph_state import GraphState
from lightcurvelynx.math_nodes.given_sampler import GivenValueList
from lightcurvelynx.models.basic_models import ConstantSEDModel, StepModel
from lightcurvelynx.models.static_sed_model import StaticBandfluxModel
from lightcurvelynx.obstable.opsim import OpSim
from lightcurvelynx.simulate import (
    compute_noise_free_lightcurves,
    compute_single_noise_free_lightcurve,
    get_time_windows,
    simulate_lightcurves,
)


def test_get_time_windows():
    """Test the get_time_windows function with various inputs."""
    assert get_time_windows(None, None) == (None, None)
    assert get_time_windows(0.0, None) == (None, None)
    assert get_time_windows(None, (1.0, 2.0)) == (None, None)

    result = get_time_windows(0.0, (1.0, 2.0))
    assert np.array_equal(result[0], np.array([-1.0]))
    assert np.array_equal(result[1], np.array([2.0]))

    result = get_time_windows(1.0, (None, 2.0))
    assert result[0] is None
    assert np.array_equal(result[1], np.array([3.0]))

    result = get_time_windows(-10.0, (1.0, None))
    assert np.array_equal(result[0], np.array([-11.0]))
    assert result[1] is None

    result = get_time_windows(np.array([0.0, 1.0, 2.0]), (1.0, 2.0))
    assert np.array_equal(result[0], np.array([-1.0, 0.0, 1.0]))
    assert np.array_equal(result[1], np.array([2.0, 3.0, 4.0]))

    with pytest.raises(ValueError):
        get_time_windows(0.0, (1.0, 2.0, 3.0))


def test_simulate_lightcurves(test_data_dir):
    """Test an end to end run of simulating the light curves."""
    # Load the OpSim data.
    opsim_db = OpSim.from_db(test_data_dir / "opsim_small.db")

    # Load the passband data for the griz filters only.
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )

    # Create a constant SED model with known brightnesses and RA, dec
    # values that match the opsim.
    given_brightness = [1000.0, 2000.0, 5000.0, 1000.0, 100.0]
    source = ConstantSEDModel(
        brightness=GivenValueList(given_brightness),
        t0=0.0,
        ra=GivenValueList(opsim_db["ra"].values[0:5]),
        dec=GivenValueList(opsim_db["dec"].values[0:5]),
        redshift=0.0,
        node_label="source",
    )

    results = simulate_lightcurves(
        source,
        5,
        opsim_db,
        passband_group,
        obstable_save_cols=["observationId", "zp_nJy"],
        param_cols=["source.brightness"],
    )
    assert len(results) == 5
    assert np.all(results["nobs"].values >= 1)
    assert np.allclose(results["ra"].values, opsim_db["ra"].values[0:5])
    assert np.allclose(results["dec"].values, opsim_db["dec"].values[0:5])
    assert np.allclose(results["z"].values, 0.0)
    assert np.allclose(results["t0"].values, 0.0)

    for idx in range(5):
        num_obs = results["nobs"][idx]
        assert len(results.loc[idx]["lightcurve"]["flux"]) == num_obs

        # Check that we pulled the metadata from the opsim.
        assert len(results.loc[idx]["lightcurve"]["observationId"]) == num_obs
        assert len(np.unique(results.loc[idx]["lightcurve"]["observationId"])) == num_obs
        assert np.all(results.loc[idx]["lightcurve"]["observationId"] >= 0)
        assert len(results.loc[idx]["lightcurve"]["zp_nJy"]) == num_obs

        # Check that we extract one of the parameters.
        assert results["source_brightness"][idx] == given_brightness[idx]

    # Check that we saved and can reassemble the GraphStates
    assert "params" in results
    state = GraphState.from_list(results["params"].values)
    assert state.num_samples == 5
    assert np.allclose(state["source.ra"], opsim_db["ra"].values[0:5])
    assert np.allclose(state["source.dec"], opsim_db["dec"].values[0:5])

    # Check that we can extract a single GraphState from the results.
    single_state = GraphState.from_dict(results["params"][2])
    assert single_state.num_samples == 1
    assert single_state["source.ra"] == opsim_db["ra"].values[2]
    assert single_state["source.dec"] == opsim_db["dec"].values[2]

    # Check that we fail if we try to save a parameter column that doesn't exist.
    # And that the error gives the existing options.
    source2 = ConstantSEDModel(
        brightness=10.0,
        t0=0.0,
        ra=1.0,
        dec=-1.0,
        redshift=0.0,
        node_label="source2",
    )
    with pytest.raises(KeyError) as excinfo:
        _ = simulate_lightcurves(
            source2,
            1,
            opsim_db,
            passband_group,
            param_cols=["source2.unknown_parameter"],
        )
    assert "Available parameters are:" in str(excinfo.value)
    assert "source2.ra" in str(excinfo.value)
    assert "source2.dec" in str(excinfo.value)


def test_simulate_bandfluxes(test_data_dir):
    """Test an end to end run of simulating a bandflux model."""
    # Create a toy observation table with two pointings.
    num_obs = 6
    obsdata = {
        "time": [0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
        "ra": [0.0, 0.0, 180.0, 0.0, 180.0, 0.0],
        "dec": [10.0, 10.0, -10.0, 10.0, -10.0, 10.0],
        "filter": ["g", "r", "g", "r", "z", "z"],
        "zp": [1.0] * num_obs,
        "seeing": [1.12] * num_obs,
        "skybrightness": [20.0] * num_obs,
        "exptime": [29.2] * num_obs,
        "nexposure": [2] * num_obs,
    }
    obstable = OpSim(obsdata)

    # Load the passband data for the griz filters only.
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )

    # Create a static bandflux model and simulate 2 runs.
    model = StaticBandfluxModel({"g": 1.0, "i": 2.0, "r": 3.0, "y": 4.0, "z": 5.0}, ra=0.0, dec=10.0)
    results = simulate_lightcurves(model, 2, obstable, passband_group)
    assert len(results) == 2
    assert np.allclose(results["lightcurve"][0]["flux_perfect"], [1.0, 3.0, 3.0, 5.0])
    assert np.allclose(results["lightcurve"][1]["flux_perfect"], [1.0, 3.0, 3.0, 5.0])


def test_simulate_single_lightcurve(test_data_dir):
    """Test an end to end run of simulating a single light curve."""
    # Load the OpSim data.
    opsim_db = OpSim.from_db(test_data_dir / "opsim_small.db")

    # Load the passband data for the griz filters only.
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )

    # Create a constant SED model with known brightnesses and RA, dec
    # values that match the opsim.
    given_brightness = [1000.0, 2000.0, 5000.0, 1000.0, 100.0]
    source = ConstantSEDModel(
        brightness=GivenValueList(given_brightness),
        t0=0.0,
        ra=GivenValueList(opsim_db["ra"].values[0:5]),
        dec=GivenValueList(opsim_db["dec"].values[0:5]),
        redshift=0.0,
        node_label="source",
    )

    results = simulate_lightcurves(
        source,
        1,
        opsim_db,
        passband_group,
        obstable_save_cols=["observationId", "zp_nJy"],
        param_cols=["source.brightness"],
    )
    assert len(results) == 1

    # Check that we saved and can reassemble the GraphStates
    assert "params" in results
    state = GraphState.from_list(results["params"].values)
    assert state.num_samples == 1
    assert state["source.ra"] == opsim_db["ra"].values[0]
    assert state["source.dec"] == opsim_db["dec"].values[0]


def test_simulate_with_time_window(test_data_dir):
    """Test an end to end run of simulating with a limited time window."""
    # Create a toy OpSim database with two pointings over a series of tiems.
    # Create a fake opsim data frame with just time, RA, and dec.
    values = {
        "observationStartMJD": np.arange(50.0),
        "fieldRA": np.array([15.0 if i % 2 == 0 else 180.0 for i in range(50)]),
        "fieldDec": np.array([10.0 if i % 2 == 0 else -10.0 for i in range(50)]),
        "filter": np.full(50, "g"),
        # We add the remaining values so the OpSim can compute noise, but they are
        # arbitrary and not tested in this test.
        "zp_nJy": np.ones(50),
        "seeingFwhmEff": np.ones(50) * 0.7,
        "visitExposureTime": np.ones(50) * 30.0,
        "numExposures": np.ones(50) * 1,
        "skyBrightness": np.full(50, 20.0),
    }
    opsim_db = OpSim(values)

    # Load the passband data for the griz filters only.
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )

    # Create a constant SED model with known brightnesses and RA, dec
    # values that match the opsim.
    source = ConstantSEDModel(
        brightness=1000.0,
        t0=GivenValueList([20.0, 15.0]),
        ra=15.0,
        dec=10.0,
        redshift=0.0,
        node_label="source",
    )

    results = simulate_lightcurves(
        source,
        2,
        opsim_db,
        passband_group,
        time_window_offset=(5.0, 10.0),
    )
    assert len(results) == 2

    # We should simulate the observations that are only within the time window, (15.0, 30.0) for the
    # first samples and (10.0, 25.0) for the second sample, and at the matching RA/Dec (the even indices).
    assert np.array_equal(
        results["lightcurve"][0]["mjd"],
        np.array([16.0, 18.0, 20.0, 22.0, 24.0, 26.0, 28.0, 30.0]),
    )
    assert np.array_equal(
        results["lightcurve"][1]["mjd"],
        np.array([10.0, 12.0, 14.0, 16.0, 18.0, 20.0, 22.0, 24.0]),
    )


def test_simulate_multiple_surveys(test_data_dir):
    """Test an end to end run of simulating a single light curve from multiple surveys."""
    # The first survey points at two locations in the sky in the "g" and "r" bands.
    obsdata1 = {
        "time": [0.0, 1.0, 2.0, 3.0],
        "ra": [0.0, 0.0, 180.0, 180.0],
        "dec": [10.0, 10.0, -10.0, -10.0],
        "filter": ["g", "r", "g", "r"],
        "zp": [0.4, 0.5, 0.6, 0.7],
        "seeing": [1.12, 1.12, 1.12, 1.12],
        "skybrightness": [20.0, 20.0, 20.0, 20.0],
        "exptime": [29.2, 29.2, 29.2, 29.2],
        "nexposure": [2, 2, 2, 2],
        "custom_col": [1, 1, 1, 1],
    }
    obstable1 = OpSim(obsdata1)
    passband_group1 = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r"],
    )

    # The second survey points at two locations on the sky in the "r" and "z" bands.
    obsdata2 = {
        "time": [0.5, 1.5, 2.5, 3.5],
        "ra": [0.0, 90.0, 0.0, 90.0],
        "dec": [10.0, -10.0, 10.0, -10.0],
        "filter": ["r", "z", "r", "z"],
        "zp": [0.05, 0.1, 0.2, 0.3],
        "seeing": [1.12, 1.12, 1.12, 1.12],
        "skybrightness": [20.0, 20.0, 20.0, 20.0],
        "exptime": [29.2, 29.2, 29.2, 29.2],
        "nexposure": [2, 2, 2, 2],
    }
    obstable2 = OpSim(obsdata2)
    passband_group2 = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["r", "z"],
    )

    # Create a constant SED model with known brightnesses and RA, dec values that
    # match the (0.0, 10.0) pointing.
    model = ConstantSEDModel(brightness=100.0, t0=0.0, ra=0.0, dec=10.0, redshift=0.0, node_label="source")
    results = simulate_lightcurves(
        model,
        1,
        [obstable1, obstable2],
        [passband_group1, passband_group2],
        obstable_save_cols=["zp", "custom_col"],
    )
    assert len(results) == 1
    assert results["nobs"][0] == 4

    # Check that the light curve was simulated correctly, including saving the zeropoint information
    # from each ObsTable.
    lightcurve = results["lightcurve"][0]
    assert np.allclose(lightcurve["mjd"], np.array([0.0, 1.0, 0.5, 2.5]))
    assert np.allclose(lightcurve["zp"], np.array([0.4, 0.5, 0.05, 0.2]))
    assert np.array_equal(lightcurve["filter"], np.array(["g", "r", "r", "r"]))
    assert np.array_equal(lightcurve["survey_idx"], np.array([0, 0, 1, 1]))

    # The custom column should only exist for observations from one of the surveys.
    assert np.all(lightcurve["custom_col"][0:2] == 1)
    assert np.all(np.isnan(lightcurve["custom_col"][2:4]))

    # We fail if we pass in lists of different lengths.
    with pytest.raises(ValueError):
        simulate_lightcurves(
            model,
            1,
            [obstable1, obstable2],
            passband_group1,
        )

    # We fail if we try to use a bandflux only model with multiple surveys.
    model2 = StaticBandfluxModel({"g": 1.0, "i": 1.0, "r": 1.0, "y": 1.0, "z": 1.0})
    with pytest.raises(ValueError):
        simulate_lightcurves(
            model2,
            1,
            [obstable1, obstable2],
            [passband_group1, passband_group2],
        )


def test_compute_noise_free_lightcurves_single(test_data_dir):
    """Test computing noise free light curves for a set of times and filters."""
    # Load the passband data for the griz filters only.
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )

    # Create a step model that changes brightness at t=10.0
    model = StepModel(brightness=100.0, t0=10.0, t1=30.0, node_label="source")
    graph_state = model.sample_parameters(num_samples=1)

    lightcurves = compute_single_noise_free_lightcurve(
        model,
        graph_state,
        passband_group,
        rest_frame_phase_min=-40.0,  # 40 days before t0
        rest_frame_phase_max=60.0,  # 60 days after t0
        rest_frame_phase_step=1.0,  # 1 sample per day
    )
    rest_phase = np.arange(-40.0, 60.0, 1.0)
    obs_times = rest_phase + 10.0
    assert np.array_equal(lightcurves["times"], obs_times)
    assert np.array_equal(lightcurves["rest_phase"], rest_phase)

    mask = (obs_times >= 10.0) & (obs_times <= 30.0)
    for filter in ["g", "r", "i", "z"]:
        assert filter in lightcurves
        bandfluxes = lightcurves[filter]
        assert len(bandfluxes) == len(obs_times)
        assert np.allclose(bandfluxes[~mask], 0.0)
        assert np.allclose(bandfluxes[mask], 100.0)


def test_compute_noise_free_lightcurves_multiple(test_data_dir):
    """Test computing noise free light curves for a set of times and filters
    and multiple sample states"""
    # Load the passband data for the griz filters only.
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )

    # Create a step model that changes brightness at t0
    start_times = [10.0, 15.0, 20.0, 25.0]
    end_times = [30.0, 30.0, 40.0, 40.0]
    model = StepModel(
        brightness=100.0,
        t0=GivenValueList(start_times),
        t1=GivenValueList(end_times),
        ra=0.0,
        dec=0.0,
        redshift=0.0,
        node_label="source",
    )
    graph_state = model.sample_parameters(num_samples=4)

    lightcurves = compute_noise_free_lightcurves(
        model,
        graph_state,
        passband_group,
        rest_frame_phase_min=-40.0,  # 40 days before t0
        rest_frame_phase_max=60.0,  # 60 days after t0
        rest_frame_phase_step=1.0,  # 1 sample per day
    )
    assert len(lightcurves) == 4

    rest_phase = np.arange(-40.0, 60.0, 1.0)
    for idx in range(4):
        assert lightcurves["id"][idx] == idx
        lc_df = lightcurves["lightcurve"][idx]
        assert np.array_equal(lc_df["rest_phase"], rest_phase)
        assert np.array_equal(lc_df["times"], rest_phase + start_times[idx])

        mask = (lc_df["times"] >= start_times[idx]) & (lc_df["times"] <= end_times[idx])
        for filter in ["g", "r", "i", "z"]:
            assert filter in lc_df
            bandfluxes = lc_df[filter]
            assert len(bandfluxes) == len(lc_df["times"])
            assert np.allclose(bandfluxes[~mask], 0.0)
            assert np.allclose(bandfluxes[mask], 100.0)



"""Functions for extrapolating flux past the end of a model's range of valid
wavelengths using flux = f(time, wavelengths).
"""

import numpy as np


class WaveExtrapolationModel:
    """The base class for the wavelength extrapolation methods."""

    def __init__(self):
        pass

    def __call__(self, last_wave, last_flux, query_waves):
        """Evaluate the extrapolation given the last valid points(s)
        and a list of new query points.

        Parameters
        ----------
        last_wave : float
            The last valid wavelength (in AA).
        last_flux : numpy.ndarray
            A length T array of the last valid flux values at each time
            at the last valid wavelength (in nJy).
        query_waves : numpy.ndarray
            A length W array of the query wavelengths (in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A T x W matrix of extrapolated values.
        """
        last_flux = np.asarray(last_flux)
        num_times = last_flux.shape[0]

        query_waves = np.asarray(query_waves)
        num_waves = query_waves.shape[0]
        return np.zeros((num_times, num_waves))


class ConstantExtrapolation(WaveExtrapolationModel):
    """Extrapolate using a constant value in nJy.

    Attributes
    ----------
    value : float
        The value (in nJy) to use for the extrapolation.
    """

    def __init__(self, value=0.0):
        super().__init__()

        if value < 0:
            raise ValueError("Extrapolation value must be positive.")
        self.value = value

    def __call__(self, last_wave, last_flux, query_waves):
        """Evaluate the extrapolation given the last valid points(s)
        and a list of new query points.

        Parameters
        ----------
        last_wave : float
            The last valid wavelength (in AA).
        last_flux : numpy.ndarray
            A length T array of the last valid flux values at each time
            at the last valid wavelength (in the model's units).
        query_waves : numpy.ndarray
            A length W array of the query wavelengths (in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A T x W matrix of extrapolated values.
        """
        last_flux = np.asarray(last_flux)
        num_times = last_flux.shape[0]

        query_waves = np.asarray(query_waves)
        num_waves = query_waves.shape[0]
        return np.full((num_times, num_waves), self.value)


class LastValueExtrapolation(WaveExtrapolationModel):
    """Extrapolate using the last valid value."""

    def __init__(self):
        super().__init__()

    def __call__(self, last_wave, last_flux, query_waves):
        """Evaluate the extrapolation given the last valid points(s)
        and a list of new query points.

        Parameters
        ----------
        last_wave : float
            The last valid wavelength (in AA).
        last_flux : numpy.ndarray
            A length T array of the last valid flux values at each time
            at the last valid wavelength (in nJy).
        query_waves : numpy.ndarray
            A length W array of the query wavelengths (in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A T x W matrix of extrapolated values (in nJy).
        """
        last_flux = np.asarray(last_flux)
        query_waves = np.asarray(query_waves)
        num_waves = query_waves.shape[0]

        return np.tile(last_flux[:, np.newaxis], (1, num_waves))


class LinearDecay(WaveExtrapolationModel):
    """Linear decay of the flux using the last valid point(s) down to zero.

    Attributes
    ----------
    decay_width : float
        The width of the decay region in Angstroms. The flux is
        linearly decreased to zero over this range.
    """

    def __init__(self, decay_width=100.0):
        super().__init__()

        if decay_width <= 0:
            raise ValueError("decay_width must be positive.")
        self.decay_width = decay_width

    def __call__(self, last_wave, last_flux, query_waves):
        """Evaluate the extrapolation given the last valid points(s)
        and a list of new query points.

        Parameters
        ----------
        last_wave : float
            The last valid wavelength (in AA).
        last_flux : numpy.ndarray
            A length T array of the last valid flux values at each time
            at the last valid wavelength (in nJy).
        query_waves : numpy.ndarray
            A length W array of the query wavelengths (in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A T x W matrix of extrapolated values (in nJy).
        """
        last_flux = np.asarray(last_flux)
        query_waves = np.asarray(query_waves)
        dist = np.abs(query_waves - last_wave)

        multiplier = np.clip(1.0 - (dist / self.decay_width), 0.0, 1.0)
        flux = last_flux[:, np.newaxis] * multiplier[np.newaxis, :]
        return flux


class ExponentialDecay(WaveExtrapolationModel):
    """Exponential decay of the flux using the last valid point(s) down to zero.

    f(t, λ) = f(t, λ_last) * exp(- rate * \\|λ - λ_last\\|)

    Attributes
    ----------
    rate : float
        The decay rate in the exponential function.
    """

    def __init__(self, rate):
        super().__init__()

        if rate < 0:
            raise ValueError("Decay rate must be zero or positive.")
        self.rate = rate

    def __call__(self, last_wave, last_flux, query_waves):
        """Evaluate the extrapolation given the last valid points(s)
        and a list of new query points.

        Parameters
        ----------
        last_wave : float
            The last valid wavelength (in AA).
        last_flux : numpy.ndarray
            A length T array of the last valid flux values at each time
            at the last valid wavelength (in nJy)
        query_waves : numpy.ndarray
            A length W array of the query wavelengths (in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A T x W matrix of extrapolated values (in nJy).
        """
        last_flux = np.asarray(last_flux)
        query_waves = np.asarray(query_waves)
        dist = np.abs(query_waves - last_wave)

        multiplier = np.exp(-self.rate * dist)
        flux = last_flux[:, np.newaxis] * multiplier[np.newaxis, :]
        return flux



"""A collection of sampled parameters from a statistic distribution.

Model parameters are random variables that are sampled together in a joint distribution
using a graph of dependencies. For example, the functions::

    f(a, b) = x
    g(c) = y
    h(x, y) = z

indicate that x depends on a and b, y depends on c, and z depends on x and y (and thus on a, b, and c
as well). These would form a graph that looks like::

    a - \\
         x -- \\
    b - /      \\
                z
    c -- y --- /

Within LightCurveLynx, variables are grouped into logical sets called nodes. The combination of node name
and variable name are used to indicate specific values, allowing us to use the same variable names
in multiple nodes. For example a node to generate samples from a Gaussian distribution may have internal
parameters called mean and scale that might take on different values depending on what the node is generating.
We could have one mean for an object's brightness and another for its positional relative to the center
of a host galaxy.
"""

import numpy as np
from astropy.io import ascii
from astropy.table import Table


class GraphState:
    """A class to hold the state(s) of the each variable for one or more samples of the random
    variables in the graph. Each entry is indexed by a combination of node's (unique) name and
    variable's name.  This allows nodes to have parameters with the same name, such as ra and dec.

    Attributes
    ----------
    states : dict
        A dictionary of dictionaries mapping the node's string and variable name to
        either a value or array of values for that parameters.
    num_samples : int
        A count of the number of samples stored in the GraphState.  If num_samples > 1, then
        all parameters are stored as arrays of length num_samples.  If num_samples == 1, then
        all parameters are stored as scalars.
    num_parameters : int
        The total number of parameters stored in a single sample within GraphState.
    fixed_vars : dict
        A dictionary mapping the node name to a set of the variable names that
        are fixed (not changed by resampling) in this GraphState instance.
    """

    def __init__(self, num_samples=1):
        if num_samples < 1:
            raise ValueError(
                f"Invalid number of samples for GraphState ({num_samples}). Must be a positive integer."
            )
        self.num_samples = num_samples
        self.num_parameters = 0
        self.states = {}
        self.fixed_vars = {}

    def __len__(self):
        return self.num_parameters

    def __next__(self):
        return next(self._iterate())

    def __iter__(self):
        return self._iterate()

    def _iterate(self):
        """Returns a single sliced state, which is a GraphState object
        with num_samples==1 and all scalar values."""
        for idx in range(self.num_samples):
            yield self.extract_single_sample(idx)

    def __contains__(self, key):
        """Check if the GraphState contains an entry.

        The key can be::

        1) the name of a node (in which case we return True if the node exists),
        2) the full name of a parameter (in which case we return True if the
           combination of node and parameter exists), or
        3) the name of a parameter in a GraphState with a single node (in which case we return True
           if the parameter exists in that node).

        Parameters
        ----------
        key : str
            The name of the entry to check.
        """
        if key in self.states:
            # Check if this is a node name.
            return True
        elif "." in key:
            # Check if this is a full name in node.param format.
            tokens = key.split(".")
            if len(tokens) != 2:
                raise KeyError(f"Invalid GraphState key: {key}")
            return tokens[0] in self.states and tokens[1] in self.states[tokens[0]]
        elif len(self.states) == 1:
            # Special case when we have only a single node stored in the graph state.
            node_state = list(self.states.values())[0]
            if key in node_state:
                return True
        else:
            return False

    def __str__(self):
        str_lines = []
        for node_name, node_vars in self.states.items():
            str_lines.append(f"{node_name}:")
            for var_name, value in node_vars.items():
                str_lines.append(f"    {var_name}: {value}")
        return "\n".join(str_lines)

    def __eq__(self, other):
        if self.num_samples != other.num_samples:
            return False
        if len(self.states) != len(other.states):
            return False

        for node_name, node_params in self.states.items():
            # Check that this node exists in both GraphStates and has the same number
            # of parameters.
            if node_name not in other.states:
                return False
            other_params = other.states[node_name]
            if len(node_params) != len(other_params):
                return False

            # Check that the values of each parameter matches.
            for var_name, var_value in node_params.items():
                if var_name not in other_params:
                    return False
                if not np.allclose(var_value, other_params[var_name]):
                    return False

        # Finally check that the 'fixed' dictionary is the same.
        return self.fixed_vars == other.fixed_vars

    def __getitem__(self, key):
        """Access an entry in the GraphState.

        The key can be::

        1) the name of a node (in which case we return that node's dictionary of
           parameter_name -> value),
        2) the full name of a parameter (in which case we return the values), or
        3) the name of a parameter in a GraphState with a single node (in which case we
           return that parameter's values).

        Parameters
        ----------
        key : str
            The name of the entry to access.
        """
        if key in self.states:
            return self.states[key]
        elif "." in key:
            tokens = key.split(".")
            if len(tokens) != 2:
                raise KeyError(f"Invalid GraphState key: {key}")
            return self.states[tokens[0]][tokens[1]]
        elif len(self.states) == 1:
            # Special case when we have only a single node stored
            # in the graph state.
            node_state = list(self.states.values())[0]
            if key in node_state:
                return node_state[key]
        else:
            raise KeyError(f"Unknown GraphState key: {key}")

    def copy(self):
        """Create a deep copy of the GraphState.

        Returns
        -------
        GraphState
            The copied GraphState.
        """
        new_state = GraphState(num_samples=self.num_samples)
        new_state.num_parameters = self.num_parameters
        for node_name, node_vars in self.states.items():
            new_state.states[node_name] = {}
            for var_name, var_value in node_vars.items():
                if self.num_samples == 1:
                    new_state.states[node_name][var_name] = var_value
                else:
                    new_state.states[node_name][var_name] = var_value.copy()
        for node_name, fixed_vars in self.fixed_vars.items():
            new_state.fixed_vars[node_name] = fixed_vars.copy()
        return new_state

    @staticmethod
    def extended_param_name(node_name, param_name):
        """A helper function to create the full parameter name.

        Parameters
        ----------
        node_name : str
            The name of the node.
        param_name : str
            The name of the parameter.

        Returns
        -------
        extended : str
            A name of the form {node_name}.{param_name}
        """
        return f"{node_name}.{param_name}"

    @classmethod
    def from_table(cls, input_table):
        """Create the GraphState from an AstroPy Table with columns for each parameter
        and column names of the form '{node_name}.{param_name}'.

        Parameters
        ----------
        input_table : astropy.table.Table
            The input table.
        """
        num_samples = len(input_table)
        result = GraphState(num_samples=num_samples)
        for col in input_table.colnames:
            components = col.split(".")
            if len(components) != 2:
                raise ValueError(
                    f"Invalid name for entry '{col}'. Entries should be of the form 'node_name.param_name'."
                )

            # If we only have a single value then store that value instead of the np array.
            if num_samples == 1:
                result.set(components[0], components[1], input_table[col].data[0])
            else:
                result.set(components[0], components[1], input_table[col].data)
        return result

    @classmethod
    def from_file(cls, filename):
        """Create the GraphState from a saved file.

        Parameters
        ----------
        filename : str or Path
            The name of the file.
        """
        data_table = ascii.read(filename, format="ecsv")
        return GraphState.from_table(data_table)

    @classmethod
    def from_dict(cls, data, num_samples=1):
        """Create a GraphState from either a flattened dictionary, where the keys of the
        dictionary are {node_name}.{param_name}, or a nested dictionary, where
        data[node_name][param_name] = value.

        Parameters
        ----------
        data : dict
            The dictionary mapping the parameter identifier (node name and parameter name)
            to their values.
        num_samples : int
            The number of samples.
            Default:  1

        Returns
        -------
        GraphState
            The corresponding graph state.
        """
        state = GraphState(num_samples=num_samples)
        for id1, val1 in data.items():
            if "." in id1:
                # Handle the flattened array by splitting the key.
                node_name, param_name = id1.split(".")
                state.set(node_name, param_name, val1, force_copy=True, fixed=False)
            elif isinstance(val1, dict):
                # Handle the nested array by iterating over the second dictionary's entries.
                for param_name, values in val1.items():
                    state.set(id1, param_name, values, force_copy=True, fixed=False)
            else:
                raise ValueError("Input dictionary must either be flattened or nested.")
        return state

    @classmethod
    def from_list(cls, data):
        """Concatenate a list of GraphStates or single state dictionaries into a single GraphState.
        All the entries in the data must have the same set of parameters (keys).

        Parameters
        ----------
        data : list of GraphState or dict
            A list of the individual GraphState information to combine.

        Returns
        -------
        GraphState
            The corresponding graph state.
        """
        if len(data) == 0:
            raise ValueError("Cannot concatenate an empty list.")

        # Convert everything into GraphStates (if they are not already) and extract
        # the basic information.
        all_param_full_names = None
        graph_states = []
        total_samples = 0
        for current in data:
            if isinstance(current, dict):
                current = GraphState.from_dict(current)
            elif not isinstance(current, GraphState):
                raise TypeError(f"Concatenate takes either GraphState or dict. Got {type(current)}")

            # Check that this is either the first GraphState we have seen or has the same parameters
            # as the earlier GraphStates we have seen.
            current_full_names = set(current.get_all_params_names())
            if all_param_full_names is None:
                all_param_full_names = current_full_names
            elif all_param_full_names != current_full_names:
                raise ValueError(
                    f"The sets of parameters do not match. Expected {all_param_full_names}."
                    f"Received {current_full_names}."
                )

            total_samples += current.num_samples
            graph_states.append(current)

        # Allocate space for the concatenated states and fill that result.
        result = GraphState(num_samples=total_samples)
        for full_name in all_param_full_names:
            node_name, param_name = full_name.split(".")

            # Create a numpy array that is the concatenation of all the values
            # from each of the GraphStates.
            values_list = []
            for current in graph_states:
                values_list.append(np.atleast_1d(current[full_name]))
            values = np.concatenate(values_list)
            result.set(node_name, param_name, values, force_copy=False, fixed=False)
        return result

    def get_all_params_names(self):
        """Get the full name of all the parameters.

        Returns
        -------
        names : list
            A list of all the parameter names.
        """
        names = []
        for node_name, params in self.states.items():
            for param_name in params:
                names.append(self.extended_param_name(node_name, param_name))
        return names

    def get_node_state(self, node_name, sample_num=0):
        """Get a dictionary of all parameters local to the given node
        for a single sample state.

        Parameters
        ----------
        node_name : str
            The parent node whose variables to extract.
        sample_num : int
            The number of sample to extract.

        Returns
        -------
        values : dict
            A dictionary mapping the parameter name to its value.
        """
        if node_name not in self.states:
            raise KeyError(f"Node name '{node_name}' not found in GraphState.")
        if sample_num < 0 or sample_num >= self.num_samples:
            raise ValueError(f"Invalid index {sample_num} in GraphState with {self.num_samples} entries.")
        if self.num_samples == 1:
            values = self.states[node_name]
        else:
            values = {}
            for var_name, val in self.states[node_name].items():
                values[var_name] = val[sample_num]
        return values

    def set(self, node_name, var_name, value, force_copy=False, fixed=False):
        """Set a (new) parameter's value(s) in the GraphState from a given constant value
        or an array of length num_samples (to set all the values at once).

        Parameters
        ----------
        node_name : str
            The parent node holding this variable.
        var_name : str
            The parameter's name.
        value : any
            The new value of the parameter.
        force_copy : bool
            Make a copy of data in an array. If set to False, this will link
            to the array, saving memory and computation time.
            Default: False
        fixed : bool
            Treat this parameter as fixed and do not change it during subsequent calls to set.
            Default: False
        """
        # Check that the names do not use the separator value.
        if "." in node_name or "." in var_name:
            raise ValueError("GraphState names (node or variable) cannot contain the character '.'")

        # Update the meta data.
        if node_name not in self.states:
            self.states[node_name] = {}
            self.fixed_vars[node_name] = set()
        if var_name not in self.states[node_name]:
            self.num_parameters += 1

        # Check if this parameter is fixed. If so, skip the set.
        if var_name in self.fixed_vars[node_name]:
            return

        # Set the actual values.
        if self.num_samples == 1:
            # If this GraphState holds only a single sample, set it from the given value.
            self.states[node_name][var_name] = value
        elif np.isscalar(value):
            # If the value is a scalar, expand it to the correct number of samples.
            self.states[node_name][var_name] = np.full(self.num_samples, value)
        elif len(value) != self.num_samples:
            raise ValueError(
                f"Incompatible number of samples when setting GraphState for node={node_name}, "
                f"variable={var_name}: {self.num_samples} vs {len(value)}."
            )
        elif force_copy:
            self.states[node_name][var_name] = np.array(value.copy())
        else:
            self.states[node_name][var_name] = np.asarray(value)

        # Mark the variable as fixed if needed.
        if fixed:
            self.fixed_vars[node_name].add(var_name)

    def update(self, inputs, force_copy=False, all_fixed=False):
        """Set multiple parameters' value in the GraphState from a GraphState or a
        dictionary of the same form.

        Note
        ----
        The number of samples in input must either match the number of samples in the
        current object or be 1.

        Parameters
        ----------
        inputs : GraphState or dict
            Values to copy.
        force_copy : bool
            Make a copy of data in an array. If set to False, this will link
            to the array, saving memory and computation time.
            Default: False
        all_fixed : bool
            Treat all the parameters in inputs as fixed.
            Default: False

        Raises
        ------
        ValueError
            If the input an invalid number of samples.
        """
        if isinstance(inputs, GraphState):
            if self.num_samples != inputs.num_samples and inputs.num_samples != 1:
                raise ValueError(
                    f"GraphStates must have the same number of samples. "
                    f"Received {self.num_samples} and {inputs.num_samples}."
                )
            new_states = inputs.states
        else:
            new_states = inputs

        # Set the values one by one. The set function takes care of expanding
        # any values that are constants (e.g. float or int) to match the correct
        # number of samples.
        for node_name, node_vars in new_states.items():
            for var_name, value in node_vars.items():
                self.set(node_name, var_name, value, force_copy=force_copy, fixed=all_fixed)

    def extract_single_sample(self, sample_num):
        """Create a new GraphState with a single sample state and all scalar values.

        Parameters
        ----------
        sample_num : int
            The number of sample to extract.
        """
        if self.num_samples <= 0:
            raise ValueError("Cannot sample an empty GraphState")
        if sample_num < 0 or sample_num >= self.num_samples:
            raise ValueError(f"Invalid index {sample_num} in GraphState with {self.num_samples} entries.")

        # Make a copy of the GraphState with exactly one sample.
        new_state = GraphState(1)
        new_state.num_parameters = self.num_parameters
        for node_name in self.states:
            new_state.states[node_name] = {}
            for var_name, value in self.states[node_name].items():
                if self.num_samples == 1:
                    new_state.states[node_name][var_name] = value
                else:
                    new_state.states[node_name][var_name] = value[sample_num]
        return new_state

    def extract_parameters(self, params):
        """Extract the parameter value(s) by a given name. This is often used for
        recording the important parameters from an entire model (set of nodes).

        Parameters
        ----------
        params : str or list-like, optional
            The parameter names to extract. These can be full names ("node.param") or
            use the parameter names.

        Returns
        -------
        values : dict
            The resulting dictionary.
        """
        # If we are looking up a single parameter, but it into a list.
        if isinstance(params, str):
            params = [params]

        # Go through all the parameters. If a parameters full name is provided,
        # look it up now and save the result. Otherwise put it into a list to check
        # for in each node.
        single_params = set()
        results = {}
        for current in params:
            if "." in current:
                node_name, param_name = current.split(".")
                if node_name in self.states and param_name in self.states[node_name]:
                    results[current] = self.states[node_name][param_name]
                else:
                    raise KeyError(f"Parameter '{current}' not found in GraphState.")
            else:
                single_params.add(current)

        if len(single_params) == 0:
            # Nothing else to do.
            return results

        # Traverse the nested dictionaries looking for cases where the parameter names match.
        first_seen_node = {}
        for node_name, node_params in self.states.items():
            for param_name, param_value in node_params.items():
                if param_name in single_params:
                    if param_name in first_seen_node:
                        # We've already seen this parameter in another node. Time to use the
                        # expanded names.

                        # Start by expanding the result we have already seen if needed.
                        if param_name in results:
                            full_name_existing = GraphState.extended_param_name(
                                first_seen_node[param_name],
                                param_name,
                            )
                            results[full_name_existing] = results[param_name]
                            del results[param_name]

                        # Add the result from the current node.
                        full_name_current = GraphState.extended_param_name(node_name, param_name)
                        results[full_name_current] = param_value
                    else:
                        # This is the first time we have seen the node. Save it with
                        # just the parameter name. Also save the node where we saw it.
                        results[param_name] = param_value
                        first_seen_node[param_name] = node_name

        # Check that we found a match for all the short parameter names.
        for param_name in single_params:
            if param_name not in first_seen_node:
                raise KeyError(f"Parameter '{param_name}' not found in GraphState.")

        return results

    def to_table(self):
        """Flatten the graph state to an AstroPy Table with columns for each parameter.

        The column names are: {node_name}.{param_name}

        Returns
        -------
        values : astropy.table.Table
            The resulting Table.
        """
        values = Table()
        for node_name, node_params in self.states.items():
            for param_name, param_value in node_params.items():
                values[self.extended_param_name(node_name, param_name)] = np.array(param_value)
        return values

    def to_dict(self):
        """Flatten the graph state to a dictionary with columns for each parameter.

        The column names are: {node_name}.{param_name}

        Returns
        -------
        values : dict
            The resulting dictionary.
        """
        values = {}
        for node_name, node_params in self.states.items():
            for param_name, param_value in node_params.items():
                if self.num_samples == 1:
                    values[self.extended_param_name(node_name, param_name)] = param_value
                else:
                    values[self.extended_param_name(node_name, param_name)] = list(param_value)
        return values

    def save_to_file(self, filename, overwrite=False):
        """Save the GraphState to a file.

        Parameters
        ----------
        filename : str
            The name of the file to save.
        overwrite : bool
            Whether to overwrite an existing file.
            Default: False
        """
        data_table = self.to_table()
        ascii.write(data_table, filename, format="ecsv", overwrite=overwrite)


class DependencyGraph:
    """A class to hold the dependencies between parameters in a model. Used for
    analysis, documentation, testing, and visualization of the model structure.
    The full parameter names are in the same form used by GraphState.

    Attributes
    ----------
    all_params : set
        A set of all (full) parameter names in the graph.
    all_nodes : set
        A set of all node names in the graph.
    incoming : dict
        A dictionary mapping each parameter to the list of parameters that it depends on
        (the incoming edges).
    outgoing : dict
        A dictionary mapping each parameter to the list of parameters that depend on it
        (the outgoing edges).
    num_constants : int
        The number of constant parameters in the graph.
    """

    def __init__(self):
        self.all_params = set()
        self.all_nodes = set()
        self.incoming = {}
        self.outgoing = {}
        self.num_constants = 0

    def __len__(self):
        return len(self.all_params)

    def __contains__(self, full_param_name):
        return full_param_name in self.all_params

    def add_parameter(self, param_name, node_name=None):
        """Add a parameter to the dependency graph if it is not already present.

        Parameters
        ----------
        param_name : str
            The name of the parameter to add.
        node_name : str, optional
            The name of the node holding this parameter. If provided, the full parameter
            name will be in the same form used by GraphState for storage.
            Default: None
        """
        # If a node name is provided, create the expanded parameter name.
        # Also add the node name to the set of all nodes.
        if node_name is not None:
            if node_name not in self.all_nodes:
                self.all_nodes.add(node_name)
            param_name = GraphState.extended_param_name(node_name, param_name)

        # If we haven't seen the parameter before, add it to the graph.
        if param_name not in self.all_params:
            self.all_params.add(param_name)
            self.incoming[param_name] = []
            self.outgoing[param_name] = []

    def add_constant(self, value):
        """Add a constant parameter to the dependency graph.

        Parameters
        ----------
        value : any
            The value of the constant.

        Returns
        -------
        const_name : str
            The name of the constant parameter added to the graph.
        """
        const_name = f"const_{self.num_constants}={value}"
        self.num_constants += 1
        self.add_parameter(const_name)
        return const_name

    def add_edge(self, from_param, to_param):
        """Add a directed edge to the dependency graph.

        Parameters
        ----------
        from_param : str
            The name of the parameter that the edge is coming from (the dependency).
        to_param : str
            The name of the parameter that the edge is going to (the dependent).
        """
        if from_param not in self.all_params or to_param not in self.all_params:
            raise KeyError("Both parameters must be added to the graph before adding an edge.")
        self.incoming[to_param].append(from_param)
        self.outgoing[from_param].append(to_param)


def transpose_dict_of_list(input_dict, num_elem):
    """Transpose a dictionary of iterables to a list of dictionaries.

    Parameters
    ----------
    input_dict : dict
        A dictionary of iterables, each of which is length num_elem.
    num_elem : int
        The length of the iterables.

    Returns
    -------
    output_list : list
        A length num_elem list of dictionaries, each with the same keys mapping
        to a single value.

    Raises
    ------
    ValueError
        If any of the iterables have different lengths.
    """
    if num_elem < 1:
        raise ValueError(f"Trying to transpose a dictionary with {num_elem} elements")

    output_list = [{} for _ in range(num_elem)]
    for key, values in input_dict.items():
        if len(values) != num_elem:
            raise ValueError(f"Entry '{key}' has length {len(values)}. Expected {num_elem}.")
        for i in range(num_elem):
            output_list[i][key] = values[i]
    return output_list



"""The top-level module for survey related data, such as pointing and noise
information. ObsTable class is a base class with specific implementations
for different survey data, such as Rubin and ZTF."""

import sqlite3
from pathlib import Path

import astropy.units as u
import numpy as np
import pandas as pd
from astropy.coordinates import Latitude, Longitude
from mocpy import MOC
from regions import Region
from scipy.spatial import KDTree

from lightcurvelynx.astro_utils.detector_footprint import DetectorFootprint


class ObsTable:
    """A wrapper class around the observations table with helper computation functions and
    cached data for efficiency.

    Parameters
    ----------
    table : dict or pandas.core.frame.DataFrame
        The table with all the survey information. Metadata can be included in the
        "lightcurvelynx_survey_data" entry of the attributes dictionary.
    colmap : dict, optional
        A mapping of standard column names to their names in the input table.
        For example, in Rubin's OpSim we might have the column "observationStartMJD"
        which maps to "time". In that case we would have an entry with key="time"
        and value="observationStartMJD".
    detector_footprint : astropy.regions.SkyRegion, Astropy.regions.PixelRegion, or
        DetectorFootprint, optional
        The footprint object for the instrument's detector. If None, no footprint
        filtering is done. Default is None.
    wcs : astropy.wcs.WCS, optional
        The WCS for the footprint. Either this or pixel_scale must be provided if
        a footprint is provided as a Astropy region.
    **kwargs : dict
        Additional keyword arguments to pass to the constructor. This can include
        overrides of any of the survey values.

    Attributes
    ----------
    survey_values : dict, optional
        A mapping for constant values for the survey used in various computations, such
        as readout noise and dark current.
    filters : np.ndarray
        The unique filters in the survey table (if provided).
    _table : pandas.core.frame.DataFrame
        The table with all the observation information mapped to standard column names.
    _colmap : dict
        A mapping of standard column names to their names in the input table.
    _inv_colmap : dict
        A dictionary mapping the custom column names back to the standard names.
    _kd_tree : scipy.spatial.KDTree or None
        A kd_tree of the survey pointings for fast spatial queries. We use the scipy
        kd-tree instead of astropy's functions so we can directly control caching.
    _detector_footprint : DetectorFootprint, optional
        The footprint object for the instrument's detector. If None, no footprint
        filtering is done. Default is None.
    _wacs : astropy.wcs.WCS, optional
        The WCS for the footprint.
    """

    _required_columns = ["ra", "dec", "time"]

    # Default survey values. These are all None for the abstract base class.
    _default_survey_values = {
        "dark_current": None,
        "ext_coeff": None,
        "pixel_scale": None,
        "radius": None,
        "read_noise": None,
        "zp_per_sec": None,
        "survey_name": "Unknown",
    }

    def __init__(
        self,
        table,
        *,
        colmap=None,
        detector_footprint=None,
        wcs=None,
        **kwargs,
    ):
        # Create a copy of the table.
        if isinstance(table, dict):
            self._table = pd.DataFrame(table)
        else:
            self._table = table.copy()

        # Remap the columns to standard names. Start with the existing names (from the table)
        # and overwrite anything provided by the column map. Save the inverse mapping.
        name_map = {col: col for col in self._table.columns}
        self._inv_colmap = {}
        self._colmap = colmap if colmap is not None else {}
        if colmap is not None:
            for key, value in colmap.items():
                if value in name_map:
                    # Check for collisions (mapping a column to an existing column)
                    if key in self._table.columns and key != value:
                        raise ValueError(f"Trying to map {value} to {key}, but {key} is already a column.")

                    # Add this entry to the list of column names that need to be remapped.
                    name_map[value] = key

                # Save the inverse mapping as well
                self._inv_colmap[value] = key
        self._table.rename(columns=name_map, inplace=True)

        # Check that we have the required columns.
        for col in self._required_columns:
            if col not in self._table.columns:
                raise KeyError(f"Missing required column: {col}")

        # Save the survey values, overwriting anything that is manually specified
        # as a keyword argument or provided in the table's metadata.
        self.survey_values = self._default_survey_values.copy()
        if "lightcurvelynx_survey_data" in self._table.attrs:
            metadata = self._table.attrs["lightcurvelynx_survey_data"]
            if not isinstance(metadata, dict):
                raise TypeError("Got unexpected type for lightcurvelynx_survey_data")
            for key, value in metadata.items():
                self.survey_values[key] = value
        for key, value in kwargs.items():
            self.survey_values[key] = value

        self.filters = np.unique(self._table["filter"]) if "filter" in self._table.columns else np.array([])

        # If we are not given zero point data, try to derive it from the other columns.
        if "zp" not in self:
            self._assign_zero_points()

        # Build the kd-tree.
        self._kd_tree = None
        self._build_kd_tree()

        # Create the footprint if one is provided.
        self._wcs = wcs
        if isinstance(detector_footprint, Region):
            pixel_scale = self.survey_values.get("pixel_scale", None)
            detector_footprint = DetectorFootprint(detector_footprint, wcs=wcs, pixel_scale=pixel_scale)
        self._detector_footprint = detector_footprint

    def __len__(self):
        return len(self._table)

    def __getitem__(self, key):
        """Access the underlying observation table by column name."""
        if key in self._table.columns:
            return self._table[key]
        if key in self._inv_colmap and self._inv_colmap[key] in self._table.columns:
            return self._table[self._inv_colmap[key]]
        raise KeyError(f"Column not found: {key}")

    def __contains__(self, key):
        """Check if a column exists in the survey table."""
        if key in self._table.columns:
            return True
        if key in self._inv_colmap and self._inv_colmap[key] in self._table.columns:
            return True
        return False

    def clear_detector_footprint(self):
        """Clear the detector footprint, so no footprint filtering is done."""
        self._detector_footprint = None

    def _assign_constant_if_needed(self, colname, check_positive=True):
        """Assign a constant column to the table if it does not already have one.

        Parameters
        ----------
        colname : str
            The name of the column to assign.
        check_positive : bool, optional
            If True, check that the value is positive before assigning it. Default is True.

        Returns
        -------
        bool
            True if the column was added or already exists. False otherwise.
        """
        if colname in self._table.columns:
            return True  # Already have a column.

        # Check that we have the value to add.
        if self.survey_values.get(colname) is None:
            return False
        value = self.survey_values[colname]

        # If the value is a single number, convert it to a dictionary with the same value for all bands.
        if isinstance(value, int | float):
            value = {fil: value for fil in self.filters}

        # Fill in the information for each filter.
        col_vals = np.zeros(len(self._table), dtype=float)
        for fil in self.filters:
            if fil not in value:
                raise ValueError(f"`{colname}` must include all the filters in the table. Missing '{fil}'.")
            if not isinstance(value[fil], int | float):
                raise ValueError(f"`{colname}` must map filter names to numeric values.")
            if check_positive and value[fil] <= 0:
                raise ValueError(f"`{colname}` values must be positive. Got {value[fil]} for filter {fil}.")
            mask = self._table["filter"] == fil
            col_vals[mask] = value[fil]
        self.add_column(colname, col_vals)

        return True

    def safe_get_survey_value(self, key):
        """Get a survey value by key, checking that it is not None.

        Parameters
        ----------
        key : str
            The key of the survey value to retrieve.
        """
        value = self.survey_values.get(key, None)
        if value is None:
            raise ValueError(
                f"Survey value for {key} is not defined. This should be set when creating the object."
            )
        return value

    @property
    def radius(self):
        """Return the radius if it exists."""
        return self.survey_values.get("radius", None)

    @radius.setter
    def radius(self, new_val):
        """Create a setter for radius."""
        if new_val <= 0:
            raise ValueError(f"Invalid radius: {new_val}")
        self.survey_values["radius"] = new_val

    @property
    def columns(self):
        """Get the column names."""
        return self._table.columns

    @classmethod
    def from_db(cls, filename, sql_query="SELECT * FROM observations", **kwargs):
        """Create an ObsTable object from the data in an db file. Reads data matching
        what is produced by write_db (and matching the RubinOpsim table).

        Parameters
        ----------
        filename : str
            The name of the db file.
        sql_query : str
            The SQL query to use when loading the table.
            Default: "SELECT * FROM observations"
        kwargs : dict, optional
            Additional keyword arguments to pass to the Survey constructor.

        Returns
        -------
        ObsTable
            A table with all of the pointing data.

        Raise
        -----
        FileNotFoundError if the file does not exist.
        ValueError if unable to load the table.
        """
        if not Path(filename).is_file():
            raise FileNotFoundError(f"db file {filename} not found.")
        con = sqlite3.connect(f"file:{filename}?mode=ro", uri=True)

        # Read the table.
        try:
            survey_data = pd.read_sql_query(sql_query, con)
        except Exception:
            raise ValueError("Database read failed.") from None

        # Close the connection.
        con.close()

        return cls(survey_data, **kwargs)

    @classmethod
    def from_parquet(cls, filename):
        """Create an ObsTable object from a parquet file.

        Parameters
        ----------
        filename : str
            The name of the parquet file to read.

        Returns
        -------
        ObsTable
            A table with all of the pointing data.
        """
        if not Path(filename).is_file():
            raise FileNotFoundError(f"File {filename} not found.")
        survey_data = pd.read_parquet(filename)
        return cls(survey_data)

    def build_moc(self, *, radius=None, max_depth=10):
        """Build a Multi-Order Coverage Map from the regions in the data set.

        Parameters
        ----------
        radius : float, optional
            The radius to use for each image (in degrees). If not provided, the default
            radius from the survey values will be used.
        max_depth : int, optional
            The maximum depth of the MOC. Default is 10.

        Returns
        -------
        MOC
            The Multi-Order Coverage Map constructed from the data set.
        """
        radius = radius if radius is not None else self.survey_values.get("radius", None)
        if radius is None:
            raise ValueError("Radius must be provided for MOC construction or as a default. Got None.")

        longitudes = Longitude(self._table["ra"].to_list(), unit="deg")
        latitudes = Latitude(self._table["dec"].to_list(), unit="deg")
        moc = MOC.from_cones(
            lon=longitudes,
            lat=latitudes,
            radius=radius * u.deg,
            max_depth=max_depth,
            delta_depth=0,
            union_strategy="large_cones",
        )
        return moc

    def _build_kd_tree(self):
        """Construct the KD-tree from the ObsTable."""
        ra_rad = np.radians(self._table["ra"].to_numpy())
        dec_rad = np.radians(self._table["dec"].to_numpy())
        # Convert the pointings to Cartesian coordinates on a unit sphere.
        x = np.cos(dec_rad) * np.cos(ra_rad)
        y = np.cos(dec_rad) * np.sin(ra_rad)
        z = np.sin(dec_rad)
        cart_coords = np.array([x, y, z]).T

        # Construct the kd-tree.
        self._kd_tree = KDTree(cart_coords)

    def _assign_zero_points(self):
        """Assign instrumental zero points in nJy to the data table.

        Default implementation does not produce a zeropoint column. Subclasses
        should override this method with a survey specific computation.
        """
        pass

    def add_column(self, colname, values, *, overwrite=False):
        """Add a column to the current data table.

        Parameters
        ----------
        colname : str
            The name of the new column.
        values : int, float, str, list, or numpy.ndarray
            The value(s) to add.
        overwrite : bool
            Overwrite the column is it already exists.
            Default: False
        """
        if colname in self._table.columns and not overwrite:
            raise KeyError(f"Column {colname} already exists.")

        # If the input is a scalar, turn it into an array of the correct length
        if np.isscalar(values):
            values = np.full((len(self._table)), values)
        self._table[colname] = values

    def write_db(self, filename, *, tablename="observations", overwrite=False):
        """Write out the observation table as a database to a given SQL table.

        Parameters
        ----------
        filename : str
            The name of the db file.
        tablename : str
            The table to which to write.
            Default: "observations"
        overwrite : bool
            Overwrite the existing DB file.
            Default: False

        Raise
        -----
        FileExistsError if the file already exists and overwrite is False.
        """
        if_exists = "replace" if overwrite else "fail"

        con = sqlite3.connect(filename)
        try:
            self._table.to_sql(tablename, con, if_exists=if_exists)
        except Exception:
            raise ValueError("Database write failed.") from None

        con.close()

    def write_parquet(self, filename, *, overwrite=False):
        """Write out the observation table as a parquet file.

        Parameters
        ----------
        filename : str
            The name of the parquet file.
        overwrite : bool
            Overwrite the existing parquet file.
            Default: False

        Raise
        -----
        FileExistsError if the file already exists and overwrite is False.
        """
        if not overwrite and Path(filename).is_file():
            raise FileExistsError(f"File {filename} already exists.")

        # Save all the survey data as metadata.
        self._table.attrs["lightcurvelynx_survey_data"] = self.survey_values
        self._table.to_parquet(filename)

    def time_bounds(self):
        """Returns the min and max times for all observations in the ObsTable.

        Returns
        -------
        t_min, t_max : float, float
            The min and max times for all observations in the ObsTable.
        """
        t_min = self._table["time"].min()
        t_max = self._table["time"].max()
        return t_min, t_max

    def filter_rows(self, rows):
        """Filter the rows in the ObsTable to only include those indices that are provided
        in a list of row indices (integers) or marked True in a mask.

        Parameters
        ----------
        rows : numpy.ndarray
            Either a Boolean array of the same length as the table or list of integer
            row indices to keep.

        Returns
        -------
        self : ObsTable
            The filtered ObsTable object.
        """
        # Check if we are dealing with a mask of a list of indices.
        rows = np.asarray(rows)
        if rows.dtype == bool:
            if len(rows) != len(self._table):
                raise ValueError(
                    f"Mask length mismatch. Expected {len(self._table)} rows, but found {len(rows)}."
                )
            mask = rows
        else:
            mask = np.full((len(self._table),), False)
            mask[rows] = True

        # Filter the rows in-place and build a new kd-tree.
        self._table = self._table[mask]
        self._kd_tree = None
        self._build_kd_tree()

        return self

    def is_observed(self, query_ra, query_dec, *, radius=None, t_min=None, t_max=None):
        """Check if the query point(s) fall within the field of view of any
        pointing in the ObsTable.

        Parameters
        ----------
        query_ra : float or numpy.ndarray
            The query right ascension (in degrees).
        query_dec : float or numpy.ndarray
            The query declination (in degrees).
        radius : float or None, optional
            The angular radius of the observation (in degrees).
        t_min : float or None, optional
            The minimum time (in MJD) for the observations to consider.
            If None, no time filtering is applied.
        t_max : float or None, optional
            The maximum time (in MJD) for the observations to consider.
            If None, no time filtering is applied.

        Returns
        -------
        seen : bool or list[bool]
            Depending on the input, this is either a single bool to indicate
            whether the query point is observed or a list of bools for an array
            of query points.
        """
        inds = self.range_search(query_ra, query_dec, radius=radius, t_min=t_min, t_max=t_max)
        if np.isscalar(query_ra):
            return len(inds) > 0
        return [len(entry) > 0 for entry in inds]

    def range_search(self, query_ra, query_dec, *, radius=None, t_min=None, t_max=None):
        """Return the indices of the pointings that fall within the field
        of view of the query point(s).

        Parameters
        ----------
        query_ra : float or numpy.ndarray
            The query right ascension (in degrees).
        query_dec : float or numpy.ndarray
            The query declination (in degrees).
        radius : float or None, optional
            The angular radius of the observation (in degrees). If None
            uses the default radius for the ObsTable.
        t_min : float, numpy.ndarray or None, optional
            The minimum time (in MJD) for the observations to consider.
            If None, no time filtering is applied.
        t_max : float, numpy.ndarray or None, optional
            The maximum time (in MJD) for the observations to consider.
            If None, no time filtering is applied.

        Returns
        -------
        inds : list[int] or list[numpy.ndarray]
            Depending on the input, this is either a list of indices for a single query point
            or a list of arrays (of indices) for an array of query points.
        """
        if query_ra is None or query_dec is None:
            raise ValueError("Query RA and dec must be provided for range search, but got None.")

        # Fallback to the preset radius if None is provided.
        radius = radius if radius is not None else self.survey_values.get("radius", None)
        if radius is None:
            raise ValueError("Radius must be provided for range search or as a default. Got None.")

        # If the points are scalars, make them into length 1 arrays.
        is_scalar = np.isscalar(query_ra) and np.isscalar(query_dec)
        query_ra = np.atleast_1d(query_ra)
        query_dec = np.atleast_1d(query_dec)

        # Confirm the query RA and Dec have the same length.
        if len(query_ra) != len(query_dec):
            raise ValueError("Query RA and Dec must have the same length.")
        if np.any(query_ra == None) or np.any(query_dec == None):  # noqa: E711
            raise ValueError("Query RA and dec cannot contain None.")

        # Transform the query point(s) to 3-d Cartesian coordinate(s).
        ra_rad = np.radians(query_ra)
        dec_rad = np.radians(query_dec)
        x = np.cos(dec_rad) * np.cos(ra_rad)
        y = np.cos(dec_rad) * np.sin(ra_rad)
        z = np.sin(dec_rad)
        cart_query = np.array([x, y, z]).T

        # Adjust the angular radius to a cartesian search radius and perform the search.
        adjusted_radius = 2.0 * np.sin(0.5 * np.radians(radius))
        inds = self._kd_tree.query_ball_point(cart_query, adjusted_radius)

        if t_min is not None or t_max is not None:
            num_queries = len(query_ra)
            times = self._table["time"].to_numpy()

            if t_min is None:
                t_min = np.full(num_queries, -np.inf)
            else:
                t_min = np.atleast_1d(t_min)
            if len(t_min) != num_queries:
                raise ValueError(f"t_min must be a scalar or an array of length {num_queries}.")

            if t_max is None:
                t_max = np.full(num_queries, np.inf)
            else:
                t_max = np.atleast_1d(t_max)
            if len(t_max) != num_queries:
                raise ValueError(f"t_max must be a scalar or an array of length {num_queries}.")

            # Run through each list of indices and filter by time. We need to do this
            # iteratively, because the lists can have different lengths.
            for idx, subinds in enumerate(inds):
                if len(subinds) == 0:
                    continue
                time_mask = (times[subinds] >= t_min[idx]) & (times[subinds] <= t_max[idx])
                inds[idx] = np.asarray(subinds)[time_mask]

        # Do a filtering step based on the detector's footprint. We do this after the range search,
        # because it is more expensive (but also more accurate).
        if self._detector_footprint is not None:
            # Extract the RA and dec of the pointings for later use.
            all_ra = self._table["ra"].to_numpy()
            all_dec = self._table["dec"].to_numpy()
            all_rot = None if "rotation" not in self._table.columns else self._table["rotation"].to_numpy()

            for idx, subinds in enumerate(inds):
                num_matches = len(subinds)
                if num_matches == 0:
                    continue  # Nothing to filter.

                match_rot = None if all_rot is None else all_rot[subinds]
                mask = self._detector_footprint.contains(
                    np.full(num_matches, query_ra[idx]),  # The RA coordinate of this query
                    np.full(num_matches, query_dec[idx]),  # The dec coordinate of this query
                    all_ra[subinds],  # The RA coordinates of the pointings (detector positions)
                    all_dec[subinds],  # The dec coordinates of the pointings (detector positions)
                    rotation=match_rot,  # The detector rotation angles (if available)
                )
                inds[idx] = np.asarray(subinds)[mask]

        # If the query was a scalar, we return a single list of indices.
        if is_scalar:
            inds = inds[0]
        return inds

    def get_observations(self, query_ra, query_dec, *, radius=None, t_min=None, t_max=None, cols=None):
        """Return the observation information when the query point falls within
        the field of view of a pointing in the ObsTable.

        Parameters
        ----------
        query_ra : float
            The query right ascension (in degrees).
        query_dec : float
            The query declination (in degrees).
        radius : float or None, optional
            The angular radius of the observation (in degrees). If None
            uses the default radius for the ObsTable.
        t_min : float or None, optional
            The minimum time (in MJD) for the observations to consider.
            If None, no time filtering is applied.
        t_max : float or None, optional
            The maximum time (in MJD) for the observations to consider.
            If None, no time filtering is applied.
        cols : list or str
            A list of the names of columns to extract or a single column name.
            If None returns all the columns.

        Returns
        -------
        results : dict
            A dictionary mapping the given column name to a numpy array of values.
        """
        neighbors = self.range_search(query_ra, query_dec, radius=radius, t_min=t_min, t_max=t_max)

        results = {}
        if cols is None:
            cols = self._table.columns.to_list()
        elif isinstance(cols, str):
            cols = [cols]

        for col in cols:
            # Allow the user to specify either the original or mapped column names,
            # by using the class accessor (__getitem__), instead of the table one.
            if col not in self:
                raise KeyError(f"Unrecognized column name {col}")
            results[col] = self[col][neighbors].to_numpy()
        return results

    def bandflux_error_point_source(self, bandflux, index):
        """Compute observational bandflux error for a point source.

        Parameters
        ----------
        bandflux : array_like of float
            Band bandflux of the point source in nJy.
        index : array_like of int
            The index of the observation in the ObsTable table.

        Returns
        -------
        flux_err : array_like of float
            Simulated bandflux noise in nJy.
        """
        raise NotImplementedError



import numpy as np
import pytest
from astropy.cosmology import FlatLambdaCDM
from citation_compass import find_in_citations
from lightcurvelynx.models.bayesn import BayesnModel


def test_bayesn_matchness(test_data_dir):
    """Test loading a bayesnModel object from a file and test we get the same
    results as the bayesian model
    """
    # We get the bayesian_flux from running the following M20 model code provided
    # by bayesian model
    # from bayesn.bayesn_model import SEDmodel
    # model = SEDmodel(load_model='M20_model',
    #                  filter_yaml='../data/passbands/LSST/filters.yaml')
    # ts =  np.linspace(-20, 50, 71) # Set of phases at which to generate spectra
    # for each object
    # ts = np.array([0,10])
    # N = 1 # Number of SNe to generate spectra for.
    # If you specify parameter values, the number of values passed
    # needs to match N e.g. if you want spectra for 10 objects,
    # specify 10 theta values (unless you want to use the same value
    # for all of them, in which case you can just pass one single value)
    # sim = model.simulate_spectrum(ts, N, dl=1000 ,del_M=0, z=0.001, mu=0,
    #                               ebv_mw=0, theta=1, AV=1, RV=3, eps=0)
    # l, spec, params = sim

    bayesian_flux = np.array(
        [
            [
                3.28274773e07,
                2.74288959e08,
                2.70277520e08,
                4.07166239e08,
                2.86161691e08,
                2.43041990e08,
                2.41895044e08,
                1.63319541e08,
                1.56384906e08,
                1.46499586e08,
                1.45467824e08,
                1.13228160e08,
                9.41477182e07,
                7.14146377e07,
                9.50177663e07,
                9.53728123e07,
            ],
            [
                4.24470160e06,
                1.04358573e08,
                1.53981951e08,
                2.60752039e08,
                1.62446512e08,
                1.50724949e08,
                1.76743433e08,
                1.01958387e08,
                5.71958998e07,
                3.92839096e07,
                3.79761607e07,
                3.22384974e07,
                3.60499142e07,
                6.20347574e07,
                7.25112817e07,
                7.51344153e07,
            ],
        ]
    )
    times = np.array([0, 10])
    wavelengths = np.array(
        [
            3003.0,
            4004.0,
            5005.0,
            6006.0,
            7007.0,
            8008.0,
            9009.0,
            10010.0,
            11011.0,
            12012.0,
            13013.0,
            14014.0,
            15015.0,
            16016.0,
            17017.0,
            18018.0,
        ]
    )

    redshift = 0.001
    cosmo = FlatLambdaCDM(H0=70, Om0=0.3)
    distmod = cosmo.distmod(redshift).value
    m_abs = -19.5
    model = BayesnModel(
        theta=1,
        Av=1,
        Rv=3,
        t0=0,
        ra=0,
        dec=0,
        redshift=redshift,
        _M20_model_path=test_data_dir / "BAYESN.M20",
        hsiao_model_path=test_data_dir / "hsiao.h5",
        Amplitude=np.power(10.0, -0.4 * (distmod + m_abs)),
    )
    flux_density = model.evaluate_sed(times, wavelengths)
    assert np.allclose(bayesian_flux, flux_density, rtol=0.1)


def test_bayesn_no_model(test_data_dir):
    """Test that we fail if using the wrong model directory."""
    dir_name = test_data_dir / "no_such_salt2_model_dir"
    with pytest.raises(FileNotFoundError):
        redshift = 0.001
        cosmo = FlatLambdaCDM(H0=70, Om0=0.3)
        distmod = cosmo.distmod(redshift).value
        m_abs = -19.5
        _ = BayesnModel(
            theta=1,
            Av=1,
            Rv=3,
            t0=0,
            ra=0,
            dec=0,
            redshift=redshift,
            hsiao_model_path=test_data_dir / "hsiao.h5",
            Amplitude=np.power(10.0, -0.4 * (distmod + m_abs)),
            _M20_model_path=dir_name,
        )


def test_bayesn_citation():
    """Test the citations for the bayesian model."""
    bayesn_citations = find_in_citations("bayesnModel")
    for citation in bayesn_citations:
        assert "Mandel S., 2020" in citation
        assert "Hsiao E. Y., 2009" in citation



"""The BicubicInterpolator is used by SALT models.

It is adapted from sncosmo's BicubicInterpolator class (but implemented in JAX):
https://github.com/sncosmo/sncosmo/blob/v2.10.1/sncosmo/salt2utils.pyx
"""

import jax.numpy as jnp
from jax import jit, vmap

from lightcurvelynx.utils.io_utils import read_grid_data


@jit
def _kernel_value(input_vals):
    """A vectorized (JAX-ized) form of the kernval method from:
    https://github.com/sncosmo/sncosmo/blob/v2.10.1/sncosmo/salt2utils.pyx

    Uses:
        A = -0.5
        B = A + 2.0 = 1.5
        C = A + 3.0 = 2.5
    """
    x = jnp.abs(input_vals)

    # Start with the case of 1.0 <= x <= 2.0.
    result = -0.5 * (-4.0 + x * (8.0 + x * (-5.0 + x)))

    # Override cases where x < 1.0
    result = jnp.where(x < 1.0, x * x * (1.5 * x - 2.5) + 1.0, result)

    # Override cases where x > 2.0
    result = jnp.where(x > 2.0, 0.0, result)
    return result


def _compute_linear(ix, iy, ax, ay, z_vals):
    """Computes the linear interpolation (used for the edges). This operation is
    broken out into a function so it can be jit compiled and vmapped.

    Parameters
    ----------
    ix, iy : int
        The location of the current point in the array of z values.
    ax, ay : float
        The current point distance relative to the respective nodes in the z array.
    z_vals : JAX Array
        An 2d matrix of z values.

    Returns
    -------
    float
        The result of the linear interpolation at this point.
    """
    return (1.0 - ax) * ((1.0 - ay) * z_vals[ix][iy] + ay * z_vals[ix][iy + 1]) + ax * (
        (1.0 - ay) * z_vals[ix + 1][iy] + ay * z_vals[ix + 1][iy + 1]
    )


def _compute_cubic(ix, iy, dx, dy, z_vals):
    """Computes the bicubic interpolation (used for the center). This operation is
    broken out into a function so it can be jit compiled and vmapped.

    Parameters
    ----------
    ix, iy : int
        The location of the current point in the array of z values.
    dx, dy : float
        The current point distance relative to the respective nodes in the z array.
    z_vals : JAX Array
        An 2d matrix of z values.

    Returns
    -------
    float
        The result of the bicubic interpolation at this point.
    """
    wx = _kernel_value(jnp.asarray([dx - 1.0, dx, dx + 1.0, dx + 2.0]))
    wy = _kernel_value(jnp.asarray([dy - 1.0, dy, dy + 1.0, dy + 2.0]))

    return (
        wx[0]
        * (
            wy[0] * z_vals[ix - 1][iy - 1]
            + wy[1] * z_vals[ix - 1][iy]
            + wy[2] * z_vals[ix - 1][iy + 1]
            + wy[3] * z_vals[ix - 1][iy + 2]
        )
        + wx[1]
        * (
            wy[0] * z_vals[ix][iy - 1]
            + wy[1] * z_vals[ix][iy]
            + wy[2] * z_vals[ix][iy + 1]
            + wy[3] * z_vals[ix][iy + 2]
        )
        + wx[2]
        * (
            wy[0] * z_vals[ix + 1][iy - 1]
            + wy[1] * z_vals[ix + 1][iy]
            + wy[2] * z_vals[ix + 1][iy + 1]
            + wy[3] * z_vals[ix + 1][iy + 2]
        )
        + wx[3]
        * (
            wy[0] * z_vals[ix + 2][iy - 1]
            + wy[1] * z_vals[ix + 2][iy]
            + wy[2] * z_vals[ix + 2][iy + 1]
            + wy[3] * z_vals[ix + 2][iy + 2]
        )
    )


@jit
def expand_to_cross_products(x_vals, y_vals):
    """Create the unraveled arrays representing the grid of points from each access.

    Parameters
    ----------
    x_vals : array-like
        A length N array of the x values.
    y_vals : array-like
        A length M array of the y values.

    Returns
    -------
    all_x : JAX array
        A length N * M array of the x values.
    all_y : JAX array
        A length N * M array of the y values.
    """
    grids = jnp.meshgrid(x_vals, y_vals, indexing="ij")
    x_all = jnp.ravel(grids[0])
    y_all = jnp.ravel(grids[1])
    return x_all, y_all


class BicubicInterpolator:
    """An object that performs bicubic interpolation over a 2-d grid.

    Parameters
    ----------
    x_vals : array-like
        The values along the x-axis of the grid. The values must be sorted
        and at regular step sizes.
    y_vals : array-like
        The values along the y-axis of the grid. The values must be sorted
        and at regular step sizes.
    z_vals : array-like
        The values along the z-axis of the grid.

    Attributes
    ----------
    x_vals : BicubicAxis
        The values along the x-axis of the grid stored with precomputed information.
    y_vals : BicubicAxis
        The values along the y-axis of the grid stored with precomputed information.
    z_vals : JAX array
        The values along the z-axis of the grid.
    """

    def __init__(self, x_vals, y_vals, z_vals):
        # Load and validate the x and y values.
        self.x_vals = BicubicAxis(jnp.asarray(x_vals))
        self.y_vals = BicubicAxis(jnp.asarray(y_vals))

        # Load an validate the z values.
        self.z_vals = jnp.asarray(z_vals)
        if len(self.z_vals.shape) != 2:
            raise ValueError(f"z values should be a 2-d array. Found shape={self.z_vals.shape}")
        if self.z_vals.shape[0] != len(self.x_vals) or self.z_vals.shape[1] != len(self.y_vals):
            raise ValueError(
                f"z values wrong shape. Expected shape=({len(self.x_vals)}, {len(self.y_vals)})."
                f" Found shape={self.z_vals.shape}."
            )
        self._compute_linear = vmap(jit(_compute_linear), in_axes=(0, 0, 0, 0, None))
        self._compute_cubic = vmap(jit(_compute_cubic), in_axes=(0, 0, 0, 0, None))

    @classmethod
    def from_grid_file(cls, filename, scale_factor=1.0):
        """Load the grid data from an ASCII file and create a BicubicInterpolator.

        Parameters
        ----------
        filename : str
            The name of the grid file.
        scale_factor : float
            A multiplicative scale factor for the z values.
            Default: 1.0
        """
        x_vals, y_vals, z_vals = read_grid_data(filename)
        z_vals *= scale_factor
        return BicubicInterpolator(x_vals, y_vals, z_vals)

    def __call__(self, x_q, y_q):
        """Evaluate the bicubic interpolation at a grid of points.

        Parameters
        ----------
        x_q : array-like
            The N-length array of x values.
        y_q : array-like
            The M-length array of y values.

        Returns
        -------
        results : jaxlib.xla_extension.ArrayImpl
            An N x M array of interpolated values for each (x, y) pair.
        """
        x_q = jnp.asarray(x_q)
        y_q = jnp.asarray(y_q)
        n_xq = len(x_q)
        n_yq = len(y_q)

        # Find the first index *before* each of the query values with
        # the n-1 index in each dimension mapped to n-2.
        ix = self.x_vals.find_indices(x_q)
        iy = self.y_vals.find_indices(y_q)
        ix_all, iy_all = expand_to_cross_products(ix, iy)

        # Compute the weights to use for linear interpolation and perform the linear interpolation.
        wx, wy = expand_to_cross_products(
            (x_q - self.x_vals.values[ix]) / (self.x_vals.values[ix + 1] - self.x_vals.values[ix]),
            (y_q - self.y_vals.values[iy]) / (self.y_vals.values[iy + 1] - self.y_vals.values[iy]),
        )
        lin_res = self._compute_linear(ix_all, iy_all, wx, wy, self.z_vals)

        # Compute the cubic kernel weights for each point and the resulting
        # cubic interpolated values.
        dx, dy = expand_to_cross_products(
            (self.x_vals.values[ix] - x_q) / (self.x_vals.values[ix + 1] - self.x_vals.values[ix]),
            (self.y_vals.values[iy] - y_q) / (self.y_vals.values[iy + 1] - self.y_vals.values[iy]),
        )
        quad_res = self._compute_cubic(ix_all, iy_all, dx, dy, self.z_vals)

        # Fill in the values from the different interpolations. Use 0.0 for anything out of
        # bounds. Use linear interpolation for anything along the edges. And use cubic
        # interpolation for the points in the middle.
        x_out, y_out = expand_to_cross_products(
            self.x_vals.out_of_bounds(x_q),
            self.y_vals.out_of_bounds(y_q),
        )
        out_of_bounds = x_out | y_out

        x_edge, y_edge = expand_to_cross_products(
            (ix == 0) | (ix > self.x_vals.num_vals - 3),
            (iy == 0) | (iy > self.y_vals.num_vals - 3),
        )
        on_edge = x_edge | y_edge

        results = jnp.where(out_of_bounds, 0.0, jnp.where(on_edge, lin_res, quad_res))
        results = results.reshape(n_xq, n_yq)
        return results


class BicubicAxis:
    """A helper class that represents values for an axis of bicubic interpolation
    with restrictions on acceptable data to match the SALT2 and SALT3 models.

    Restrictions include:
    - Data must contain at least 3 values.
    - Data must be sorted.
    - Data must be spaced at regular steps.

    Attributes
    ----------
    values : JAX Array
        The values of the range.
    min_val : float
        The starting value of the range.
    max_val : float
        The maximum value of the range.
    num_vals : int
        The number of values in the range.
    regular_steps : bool
        Indicates whether the axis uses regularly sized steps.
    step_size : float
        The step size of the range.
    """

    def __init__(self, values):
        # Load and validate the y values.
        self.values = jnp.asarray(values)
        if len(self.values.shape) != 1:
            raise ValueError(
                f"The RegularRange values should be a 1-d array. Found shape={self.values.shape}."
            )
        if len(self.values) < 3:
            raise ValueError(
                f"Insufficient points for RegularRange. Required >= 3. Found {len(self.values)}."
            )
        if jnp.any(self.values[:-1] >= self.values[1:]):
            raise ValueError("The RegularRange values must be in sorted order.")
        self.num_vals = len(self.values)
        self.min_val = self.values[0]
        self.max_val = self.values[-1]

        # Check the step sizes are regular (allowing for inprecision when things are written as
        # floats). If it is regular we can use a faster algorithm to find the indices.
        # We pick and save a single core search function, so JAX does not encounter an "if" during
        # tracing (at the cost of another function call during find_indices).
        step_sizes = self.values[1:] - self.values[:-1]
        min_step = jnp.min(step_sizes)
        max_step = jnp.max(step_sizes)
        if max_step - min_step > 1e-6:
            self.regular_steps = False
            self.step_size = None
            self._initial_search = self._search_nonregular
        else:
            self.regular_steps = True
            self.step_size = 0.5 * min_step + 0.5 * max_step
            self._initial_search = self._search_regular

    def __len__(self):
        return self.num_vals

    def __str__(self):
        return f"BicubicAxis [{self.min_val},{self.max_val}]. step={self.step_size}. size={self.num_vals}"

    def _search_regular(self, query_pts):
        """Do an initial search for the query points indices given regular steps."""
        return jnp.floor((query_pts - self.min_val) / self.step_size).astype(int)

    def _search_nonregular(self, query_pts):
        """Do an initial search for the query points indices given non-regular steps."""
        return jnp.searchsorted(self.values, query_pts, side="right") - 1

    def find_indices(self, query_pts):
        """Finds the first index *before* each of the query values with
        the n - 1 index in each dimension mapped to n - 2.

        values[idx[i]] <= query_pts[i] < values[idx[i] + 1]
        for all i where idx[i] > 0 and idx[i] < n - 2.

        Parameters
        ----------
        query_pts : array-like
            The values of the query points.

        Returns
        -------
        idx : JAX Array
            A pair of arrays with the indices.
        """
        query_pts = jnp.asarray(query_pts)
        idx = self._initial_search(query_pts)
        idx = jnp.where(idx < 0, 0, idx)
        idx = jnp.where(idx >= self.num_vals - 1, self.num_vals - 2, idx)

        return idx

    def out_of_bounds(self, values):
        """Compute a Boolean array of the values that are out of bounds.

        Parameters
        ----------
        values : array-like
            The N values to test

        Returns
        -------
        results : JAX array
            A length N array indicating whether each element is out of bounds.
        """
        values = jnp.asarray(values)
        return (values < self.min_val) | (values > self.max_val)



from lightcurvelynx.models.snia_host import SNIaHost


def test_snia_host():
    """
    Test that we can initialize a host with a hostmass parameter
    """

    host = SNIaHost(hostmass=10.0)
    state = host.sample_parameters()
    assert host.get_param(state, "hostmass") == 10.0



import numpy as np
import pandas as pd
import pytest
from lightcurvelynx.obstable.ztf_obstable import ZTFObsTable, create_random_ztf_obs_data


def test_ztf_obstable_init():
    """Test initializing ZTFObsTable."""
    survey_data_table = create_random_ztf_obs_data(100)
    survey_data = ZTFObsTable(table=survey_data_table)

    assert "zp" in survey_data
    assert "time" in survey_data

    # We have all the attributes set at their default values.
    assert survey_data.survey_values["dark_current"] == 0.0
    assert survey_data.survey_values["gain"] == 6.2
    assert survey_data.survey_values["pixel_scale"] == 1.01
    assert survey_data.survey_values["radius"] == 2.735
    assert survey_data.survey_values["read_noise"] == 8
    assert survey_data.survey_values["survey_name"] == "ZTF"


def test_create_ztf_obstable_override():
    """Test that we can override the default survey values."""
    survey_data_table = create_random_ztf_obs_data(100)

    survey_data = ZTFObsTable(
        table=survey_data_table,
        dark_current=0.1,
        gain=7.1,
        pixel_scale=0.1,
        radius=1.0,
        read_noise=5.0,
    )

    # We have all the attributes set at their default values.
    assert survey_data.survey_values["dark_current"] == 0.1
    assert survey_data.survey_values["gain"] == 7.1
    assert survey_data.survey_values["pixel_scale"] == 0.1
    assert survey_data.survey_values["radius"] == 1.0
    assert survey_data.survey_values["read_noise"] == 5.0


def test_create_ztf_obstable_no_zp():
    """Create an survey_data without a zeropoint column."""
    dates = [
        "2020-01-01 12:00:00.000",
        "2020-01-02 12:00:00.000",
        "2020-01-03 12:00:00.000",
        "2020-01-04 12:00:00.000",
        "2020-01-05 12:00:00.000",
    ]
    values = {
        "obsdate": dates,
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
    }

    # We fail if we do not have the other columns needed:
    # "maglim", "sky", "fwhm", "exptime"
    with pytest.raises(ValueError):
        _ = ZTFObsTable(values)

    values["exptime"] = 0.005 * np.ones(5)
    values["maglim"] = 20.0 * np.ones(5)
    values["scibckgnd"] = np.ones(5)
    values["fwhm"] = 2.3 * np.ones(5)
    survey_data = ZTFObsTable(values)

    assert "zp" in survey_data
    assert np.all(survey_data["zp"] >= 0.0)


def test_noise_calculation():
    """Test that the noise calculation is in the right range."""
    mag = np.array([19.0])
    expected_magerr = np.array([0.1])

    flux_nJy = np.power(10.0, -0.4 * (mag - 31.4))
    survey_data = ZTFObsTable(
        table=pd.DataFrame(
            {
                "ra": 0.0,
                "dec": 0.0,
                "scibckgnd": 200.0,
                "maglim": 20.0,
                "fwhm": 2.3,
                "exptime": 30.0,
                "obsdate": "2020-01-01 12:00:00.000",
            },
            index=[0],
        )
    )
    fluxerr_nJy = survey_data.bandflux_error_point_source(flux_nJy, 0)
    magerr = 1.086 * fluxerr_nJy / flux_nJy

    np.testing.assert_allclose(magerr, expected_magerr, rtol=0.2)



import h5py
import numpy as np
from astropy import units as u
from citation_compass import CiteClass

from lightcurvelynx import _LIGHTCURVELYNX_BASE_DATA_DIR
from lightcurvelynx.astro_utils.unit_utils import flam_to_fnu
from lightcurvelynx.effects.extinction import ExtinctionEffect
from lightcurvelynx.models.physical_model import SEDModel


class BayesnModel(SEDModel, CiteClass):
    """A bayesian model for supernova type Ia

    The model is defined in (Mandel et al 2022) as:

    flux(time, wave) = H_grid * 10 ** (-0.4 * W_grid) * 10 ** (-0.4 * (distmod _ m_abs))

    This class is based on the bayesian implementation at:
    https://github.com/bayesn/bayesn/blob/main/bayesn/bayesn_model.py

    Parameterized values include:
        * ra - The object's right ascension in degrees. [from BasePhysicalModel]
        * dec - The object's declination in degrees. [from BasePhysicalModel]
        * redshift - The object's redshift. [from BasePhysicalModel]
        * t0 - The t0 of the zero phase, date. [from BasePhysicalModel]
        * Amplitude - The fixed normalisation factor for distance modulus. [from Amplitude class]
        * theta - The bayeSN theta parameter.
        * Av - The bayeSN Av parameter.
        * Rv - The bayeSN Rv parameter.

    References
    ----------
    * BayeSN: Mandel S., 2020 - https://arxiv.org/pdf/2008.07538
    * M20 model: Mandel et al. 2022 (MNRAS 510, 3, 3939-3966)
    * T21 model: Thorp et al. 2021 (MNRAS 508, 3, 4310-4331) - currently unused
    * W22 model: Ward et al. 2023 (ApJ 956, 2, 111) - currently unused
    * Hsiao spectral template: Hsiao E. Y., 2009 - https://arxiv.org/abs/1503.02293

    Attributes
    ----------
    _hsiao_phase: numpy.ndarray
        The phase for hsiao template.
    _hsiao_wave: numpy.ndarray
        The wavelengths for hsiao template.
    _hsiao_flux: numpy.ndarray
        The baseline mean intrinsic SED provided by hsiao template.
    _W0_: numpy.ndarray
        The global W0 matrix.
    _W1_: numpy.ndarray
        The global W1 matrix.
    _l_knots_: numpy.ndarray
        The interpolation knots for wavelengths.
    _tau_knots_: numpy.ndarray
        The interpolation knots for phase.

    Parameters
    ----------
    theta: parameter
        The bayeSN theta parameter.
    Av: parameter
        The bayeSN Av parameter.
        Set of host extinction values for each SN.
    Rv: parameter
        The bayeSN Rv parameter.
        Rv value for host extinction.
    Amplitude: parameter
        The distance modulus (m - M)
        Normalized to have a absolute magnitude of -19.5
    _M20_model_path: str
        The path for the M20 model file directory from the bayesian model
        Default: "bayesn-model-files/BAYESN.M20"
    W0_filename : str
        The file name of the W0 matrix.
        Default: "W0.txt"
    W1_filename: str
        The file name of the W1 matrix.
        Default: "W1.txt"
    l_knots_filename
        The file name of the knot values of wavelegnths for interpolation
        Default: "l_knots.txt"
    tau_knots_filename: str
        The file name of the knot values of times for interpolation
        Default: "tau_knots.txt"
    hsiao_model_path: str
        The path for the hsiao model template file directory.
        Default: "bayesn-model-files/hsiao.h5"
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    # A class variable for the units so we are not computing them each time.
    _FLAM_UNIT = u.erg / u.second / u.cm**2 / u.AA

    def __init__(
        self,
        theta=None,
        Av=None,
        Rv=None,
        t0=0.0,
        Amplitude=1.0,
        _M20_model_path=_LIGHTCURVELYNX_BASE_DATA_DIR / "bayesn-model-files/BAYESN.M20",
        W0_filename="W0.txt",
        W1_filename="W1.txt",
        l_knots_filename="l_knots.txt",
        tau_knots_filename="tau_knots.txt",
        hsiao_model_path=_LIGHTCURVELYNX_BASE_DATA_DIR / "bayesn-model-files/hsiao.h5",
        **kwargs,
    ):
        super().__init__(t0=t0, **kwargs)

        # define model specific parameters.
        self.add_parameter("theta", theta, **kwargs)
        self.add_parameter("Av", Av, **kwargs)
        self.add_parameter("Rv", Rv, **kwargs)
        self.add_parameter("Amplitude", Amplitude, **kwargs)

        # load the data files.
        self._W0_ = np.loadtxt(_M20_model_path / W0_filename)
        self._W1_ = np.loadtxt(_M20_model_path / W1_filename)
        self._l_knots_ = np.loadtxt(_M20_model_path / l_knots_filename)
        self._tau_knots_ = np.loadtxt(_M20_model_path / tau_knots_filename)
        with h5py.File(hsiao_model_path, "r") as file:
            data = file["default"]
            self._hsiao_phase = data["phase"][()].astype("float64")
            self._hsiao_wave = data["wave"][()].astype("float64")
            self._hsiao_flux = data["flux"][()].astype("float64")

    # HELPER FUNCTIONs:
    def compute_invkd(self, x):
        """
        Compute the operator matrix K^{-1}D to get second derivatives for natural cubic spline.

        Parameters
        ----------
        x : (n,) array_like
            Knot positions (non-uniform, strictly increasing).

        Returns
        -------
        invKD : (n, n) ndarray
            Matrix such that M = invKD @ y gives second derivatives of y.
        """
        n = len(x)
        K = np.zeros((n - 2, n - 2))  # Tridiagonal matrix from spline equation
        D = np.zeros((n - 2, n))  # Derivative matrix

        # Construct tridiagonal matrix K
        for j in range(1, n - 1):
            i = j - 1
            h0 = x[j] - x[j - 1]
            h1 = x[j + 1] - x[j]
            if i > 0:
                K[i, i - 1] = h0 / 6  # Subdiagonal
            K[i, i] = (h0 + h1) / 3  # Main diagonal
            if i < n - 3:
                K[i, i + 1] = h1 / 6  # Superdiagonal

        # Construct matrix D for computing second derivatives
        for j in range(1, n - 1):
            i = j - 1
            h0 = x[j] - x[j - 1]
            h1 = x[j + 1] - x[j]
            D[i, j - 1] = 1 / h0
            D[i, j] = -1 / h0 - 1 / h1
            D[i, j + 1] = 1 / h1

        # SOlve linear system to get inverse matrix of K times D
        invKD = np.zeros((n, n))
        invKD[1:-1, :] = np.linalg.solve(K, D)
        return invKD

    def natural_cubic_spline_basis_matrix_from_invkd(self, x, xq_array, invKD):
        """
        Compute basis matrix J for multiple query points using precomputed second derivative matrix.

        Parameters
        ----------
        x : (n,) array_like
            Knot positions.
        xq_array : (m,) array_like
            Query points.
        invKD : (n, n) array_like
            Precomputed matrix such that second_derivatives = invKD @ y.

        Returns
        -------
        J : (m, n) ndarray
            Basis matrix. Each row J[i, :] is the spline basis vector for xq_array[i].
        """
        x = np.asarray(x)
        xq_array = np.asarray(xq_array)
        n = len(x)
        m = len(xq_array)

        J = np.zeros((m, n))

        # Find indices of intervals for each query point
        idxs = np.searchsorted(x, xq_array) - 1
        idxs = np.clip(idxs, 0, n - 2)

        # Distances and spline weights
        x0 = x[idxs]
        x1 = x[idxs + 1]
        h = x1 - x0
        a = (x1 - xq_array) / h
        b = 1 - a
        c = ((a**3 - a) * h**2) / 6
        d = ((b**3 - b) * h**2) / 6

        # Fill the linear part of the basis matrix
        rows = np.arange(m)
        J[rows, idxs] = a
        J[rows, idxs + 1] = b

        # Add contribution from second derivatives
        J += c[:, None] * invKD[idxs] + d[:, None] * invKD[idxs + 1]
        return J

    def compute_second_derivatives_1d(self, x, y):
        """
        Compute natural cubic spline second derivatives (M) for 1D input.

        Parameters
        ----------
        x : (n,) array_like
            Knot positions.
        y : (n,) array_like
            Values at knots.

        Returns
        -------
        M : (n,) ndarray
            Second derivatives at knots.
        """
        n = len(x)
        h = np.diff(x)

        # Step1: Compute RHS alpha for the linear system
        alpha = np.zeros(n)
        for i in range(1, n - 1):
            alpha[i] = (3 / h[i]) * (y[i + 1] - y[i]) - (3 / h[i - 1]) * (y[i] - y[i - 1])

        # Step 2: Forward elimination (Thomas algorithm for tridiagonal system)
        wave = np.ones(n)
        mu = np.zeros(n)
        z = np.zeros(n)
        for i in range(1, n - 1):
            wave[i] = 2 * (x[i + 1] - x[i - 1]) - h[i - 1] * mu[i - 1]
            mu[i] = h[i] / wave[i]
            z[i] = (alpha[i] - h[i - 1] * z[i - 1]) / wave[i]

        # Step 3: Back substitution to solve for second derivatives
        M = np.zeros(n)
        for j in range(n - 2, 0, -1):
            M[j] = z[j] - mu[j] * M[j + 1]
        return M

    def compute_2d_second_derivatives(self, x, y, z):
        """
        Compute second derivatives in both x and y directions for 2D natural cubic spline.

        Parameters
        ----------
        x : (n,) array_like
            Knot positions in x.
        y : (m,) array_like
            Knot positions in y.
        z : (n, m) array_like
            Function values on 2D grid.

        Returns
        -------
        Mx : (n, m) ndarray
            Second derivatives with respect to x.
        My : (n, m) ndarray
            Second derivatives with respect to y.
        """
        n, m = len(x), len(y)
        Mx = np.zeros((n, m))
        My = np.zeros((n, m))

        # Compute second derivatives in x-direction (columns)
        for j in range(m):
            Mx[:, j] = self.compute_second_derivatives_1d(x, z[:, j])

        # Compute second derivatives in y-direction (rows)
        for i in range(n):
            My[i, :] = self.compute_second_derivatives_1d(y, z[i, :])

        return Mx, My

    def evaluate_natural_spline_2d_vectorized(self, x, y, z, Mx, My, xq, yq):
        """
        Vectorized 2D natural cubic spline evaluation using second derivatives.

        Parameters
        ----------
        x : (n,) array_like
            Knot positions in x.
        y : (m,) array_like
            Knot positions in y.
        z : (n, m) array_like
            Function values on 2D grid.
        Mx : (n, m) array_like
            Second derivatives w.r.t. x.
        My : (n, m) array_like
            Second derivatives w.r.t. y.
        xq : (p,) array_like
            Query positions in x.
        yq : (q,) array_like
            Query positions in y.

        Returns
        -------
        Zq : (p, q) ndarray
            Interpolated 2D surface values.
        """
        xq = np.atleast_1d(xq)
        yq = np.atleast_1d(yq)
        nx, ny = len(x), len(y)
        px, py = len(xq), len(yq)

        xi_idx = np.clip(np.searchsorted(x, xq) - 1, 0, nx - 2)
        yj_idx = np.clip(np.searchsorted(y, yq) - 1, 0, ny - 2)

        Zq = np.zeros((px, py))

        # Loop over y query points
        for j in range(py):
            y0_idx = yj_idx[j]
            y1_idx = y0_idx + 1
            hy = y[y1_idx] - y[y0_idx]
            ay = (y[y1_idx] - yq[j]) / hy
            by = 1 - ay

            f_y0 = []  # Interpolated values along x at y0
            f_y1 = []  # Interpolated values along x at y1

            # Loop over x query points
            for i in range(px):
                x0_idx = xi_idx[i]
                x1_idx = x0_idx + 1
                hx = x[x1_idx] - x[x0_idx]
                ax = (x[x1_idx] - xq[i]) / hx
                bx = 1 - ax

                # Values at 4 corners
                z00 = z[x0_idx, y0_idx]
                z10 = z[x1_idx, y0_idx]
                z01 = z[x0_idx, y1_idx]
                z11 = z[x1_idx, y1_idx]

                Mx00 = Mx[x0_idx, y0_idx]
                Mx10 = Mx[x1_idx, y0_idx]
                Mx01 = Mx[x0_idx, y1_idx]
                Mx11 = Mx[x1_idx, y1_idx]

                # Interpolate in x for both fixed y0 and y1
                fx0 = (ax**3 * Mx00 + bx**3 * Mx10) * hx / 6 + (
                    ax * (z00 - Mx00 * hx**2 / 6) + bx * (z10 - Mx10 * hx**2 / 6)
                )
                fx1 = (ax**3 * Mx01 + bx**3 * Mx11) * hx / 6 + (
                    ax * (z01 - Mx01 * hx**2 / 6) + bx * (z11 - Mx11 * hx**2 / 6)
                )

                f_y0.append(fx0)
                f_y1.append(fx1)

            # Interpolate final result in y
            f_y0 = np.array(f_y0)
            f_y1 = np.array(f_y1)

            # Spline in y using fx0 and fx1
            My0 = My[xi_idx, y0_idx]
            My1 = My[xi_idx, y1_idx]

            Zq[:, j] = (ay**3 * My0 + by**3 * My1) * hy / 6 + (
                ay * (f_y0 - My0 * hy**2 / 6) + by * (f_y1 - My1 * hy**2 / 6)
            )

        return Zq

    def evaluate_2d_cubic_spline(self, x, y, z, xq, yq):
        """
        Evaluate 2D natural cubic spline at given query points with auto second derivative computation.

        Parameters
        ----------
        x : (n,) array_like
            Knot positions in x direction.
        y : (m,) array_like
            Knot positions in y direction.
        z : (n, m) array_like
            Function values at grid points.
        xq : (p,) array_like
            Query points in x.
        yq : (q,) array_like
            Query points in y.

        Returns
        -------
        Zq : (p, q) ndarray
            Interpolated values at query grid.
        """
        # Automatically compute second derivatives in both directions
        Mx, My = self.compute_2d_second_derivatives(x, y, z)

        # Evaluate spling using the precomputed derivatives
        return self.evaluate_natural_spline_2d_vectorized(x, y, z, Mx, My, xq, yq)

    def mask_by_time(self, times, graph_state=None):
        """Compute a mask for whether a given time is of interest for a given object.
        For example, a user can use this function to generate a mask to include
        only the observations of interest for a window around the supernova.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        graph_state : GraphState, optional
            An object mapping graph parameters to their values.

        Returns
        -------
        time_mask : numpy.ndarray
            A length T array of Booleans indicating whether the time is of interest.
        """
        if graph_state is None:
            raise ValueError("graph_state needed to compute mask_by_time")

        z = self.get_param(graph_state, "redshift", 0.0)
        if z is None:
            z = 0.0

        t0 = self.get_param(graph_state, "t0", 0.0)
        if t0 is None:
            t0 = 0.0

        # Compute the mask.
        good_times = (times > t0 + -20.0 * (1.0 + z)) & (times < t0 + 50.0 * (1.0 + z))
        return good_times

    # MAIN FUNCTION:
    def compute_sed(self, times, wavelengths, graph_state, **kwargs):
        """Draw effect-free observations for this object.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps in MJD.
        wavelengths : numpy.ndarray, optional
            A length N array of rest frame wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.

        **kwargs : dict, optional
            Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """

        W_0 = self._W0_
        W_1 = self._W1_
        l_knots = self._l_knots_
        tau_knots = self._tau_knots_
        hsiao_wave = self._hsiao_wave
        hsiao_phase = self._hsiao_phase
        hsiao_flux = self._hsiao_flux
        params = self.get_local_params(graph_state)

        tau = times - params["t0"]
        W = W_0 + params["theta"] * W_1
        invKD_l = self.compute_invkd(l_knots)
        J_l = self.natural_cubic_spline_basis_matrix_from_invkd(l_knots, wavelengths, invKD_l)
        invKD_tau = self.compute_invkd(tau_knots)
        J_t = self.natural_cubic_spline_basis_matrix_from_invkd(tau_knots, tau, invKD_tau)
        J_t_T = np.atleast_2d(J_t).T
        WJt = np.matmul(W, J_t_T)
        W_grid = np.matmul(J_l, WJt)
        W_grid = np.atleast_2d(W_grid).T
        H_grid = self.evaluate_2d_cubic_spline(hsiao_phase, hsiao_wave, hsiao_flux, tau, wavelengths)
        flux_density = H_grid * 10 ** (-0.4 * W_grid)

        # Apply dust extinction law
        # Get ebv such that ebv = Rv/Av
        ebv = params["Av"] / params["Rv"]
        ext = ExtinctionEffect(extinction_model="F99", ebv=ebv, frame="rest")
        flux_density = ext.apply(flux_density, tau, wavelengths, ebv)

        # Apply the fixed distance modulus normalisation factor effect
        flux_density = flux_density * params["Amplitude"]

        # Convert to the correct units.
        flux_density = flam_to_fnu(
            flux_density,
            wavelengths,
            wave_unit=u.AA,
            flam_unit=self._FLAM_UNIT,
            fnu_unit=u.nJy,
        )

        return flux_density



"""Nodes that perform basic math operations that can be specified as strings.

The goal of this library is to save users from needing to create a bunch of
small FunctionNodes to perform basic math.
"""

import ast

# Disable unused import because we need all of these imported
# so they can be used during evaluation of the node.
import math  # noqa: F401

import jax.numpy as jnp  # noqa: F401
import numpy as np  # noqa: F401

from lightcurvelynx.base_models import FunctionNode


class BasicMathNode(FunctionNode):
    """A node that evaluates basic mathematical functions.

    The BasicMathNode wraps Python's eval() function to sanitize the input string
    and thus prevent the execution of arbitrary code. It also allows the user to write
    the expression once and execute using math, numpy, or JAX. The names of the
    variables in the expression must match the input variables provided by kwargs.

    Example:
        my_node = BasicMathNode(
            "redshift + 10.0 * sin(phase)",
            redshift=host.redshift,
            phase=source.phase,
        )

    Attributes
    ----------
    expression : str
        The expression to evaluate.
    backend : str
        The math libary to use. Must be one of: math, numpy, or jax.

    Parameters
    ----------
    expression : str
        The expression to evaluate.
    backend : str
        The math libary to use. Must be one of: math, numpy, or jax.
    node_label : str, optional
        An identifier (or name) for the current node.
    **kwargs : dict, optional
        Any additional keyword arguments. Every variable in the expression
        must be included as a kwarg.
    """

    # A list of supported Python operations. Used to prevent eval from
    # running arbitrary python expressions. The Call and Name types are special
    # cased so we can do checks and translations.
    _supported_ast_nodes = (
        ast.Module,  # Top level object when parsed as exec.
        ast.Expression,  # Top level object when parsed as eval.
        ast.Expr,  # Math expressions.
        ast.Constant,  # Constant values.
        ast.Load,  # Load a variable - must come from an approved function or variable.
        ast.Store,  # Store value - must come from an approved function or variable.
        ast.BinOp,  # Binary operations
        ast.Add,
        ast.Sub,
        ast.Mult,
        ast.Div,
        ast.FloorDiv,
        ast.Mod,
        ast.Pow,
        ast.UnaryOp,  # Uninary operations
        ast.UAdd,
        ast.USub,
        ast.Invert,
    )

    # A map from a very limited set of supported math constant/function names to
    # the corresponding names in [math, numpy, jax]. This is needed because
    # a very few functions have different names in different libraries.
    _math_map = {
        "abs": ["abs", "np.abs", "jnp.abs"],  # Special handling for math.
        "acos": ["math.acos", "np.acos", "jnp.acos"],
        "acosh": ["math.acosh", "np.acosh", "jnp.acosh"],
        "asin": ["math.asin", "np.asin", "jnp.asin"],
        "asinh": ["math.asinh", "np.asinh", "jnp.asinh"],
        "atan": ["math.atan", "np.atan", "jnp.atan"],
        "atan2": ["math.atan2", "np.atan2", "jnp.atan2"],
        "cos": ["math.cos", "np.cos", "jnp.cos"],
        "cosh": ["math.cosh", "np.cosh", "jnp.cosh"],
        "ceil": ["math.ceil", "np.ceil", "jnp.ceil"],
        "degrees": ["math.degrees", "np.degrees", "jnp.degrees"],
        "deg2rad": ["math.radians", "np.deg2rad", "jnp.deg2rad"],  # Special handling for math
        "e": ["math.e", "np.e", "jnp.e"],
        "exp": ["math.exp", "np.exp", "jnp.exp"],
        "fabs": ["math.fabs", "np.fabs", "jnp.fabs"],
        "floor": ["math.floor", "np.floor", "jnp.floor"],
        "log": ["math.log", "np.log", "jnp.log"],
        "log10": ["math.log10", "np.log10", "jnp.log10"],
        "log2": ["math.log2", "np.log2", "jnp.log2"],
        "max": ["max", "np.max", "jnp.max"],  # Special handling for math
        "min": ["min", "np.min", "jnp.min"],  # Special handling for math
        "pi": ["math.pi", "np.pi", "jnp.pi"],
        "pow": ["math.pow", "np.power", "jnp.power"],  # Special handling for numpy
        "power": ["math.pow", "np.power", "jnp.power"],  # Special handling for math
        "radians": ["math.radians", "np.radians", "jnp.radians"],
        "rad2deg": ["math.degrees", "np.rad2deg", "jnp.rad2deg"],  # Special handling for math
        "sin": ["math.sin", "np.sin", "jnp.sin"],
        "sinh": ["math.sinh", "np.sinh", "jnp.sinh"],
        "sqrt": ["math.sqrt", "np.sqrt", "jnp.sqrt"],
        "tan": ["math.tan", "np.tan", "jnp.tan"],
        "tanh": ["math.tanh", "np.tanh", "jnp.tanh"],
        "trunc": ["math.trunc", "np.trunc", "jnp.trunc"],
    }

    def __init__(self, expression, backend="numpy", node_label=None, **kwargs):
        if backend not in ["jax", "math", "numpy"]:
            raise ValueError(f"Unsupported math backend {backend}")
        self.backend = backend

        # Check the expression is pure math and translate it into the correct backend.
        self.expression = expression
        self._prepare(**kwargs)

        # Create a function from the expression. Note the expression has
        # already been sanitized and validated via _prepare().
        def eval_func(**kwargs):
            params = self.prepare_params(**kwargs)
            try:
                return eval(self.expression, globals(), params)
            except Exception as problem:
                # Provide more detailed logging, including the expression and parameters
                # used, when we encounter a math error like divide by zero.
                new_message = f"Error during math operation '{self.expression}' with args={kwargs}"
                raise type(problem)(new_message) from problem

        super().__init__(eval_func, node_label=node_label, **kwargs)

    def eval(self, **kwargs):
        """Evaluate the expression."""
        params = self.prepare_params(**kwargs)
        return eval(self.expression, globals(), params)

    @staticmethod
    def list_functions():
        """Return a list of the support functions.

        Returns
        -------
        list
            A list of the supported functions.
        """
        return list(BasicMathNode._math_map.keys())

    def prepare_params(self, **kwargs):
        """Convert all of the incoming parameters into the correct type,
        such as numpy arrays.

        Parameters
        ----------
        **kwargs : dict, optional
            The keyword arguments, including every variable in the expression.

        Returns
        -------
        params : dict
            The converted list of parameters.
        """
        params = {}
        for name, value in kwargs.items():
            if self.backend == "numpy":
                params[name] = np.array(value)
            elif self.backend == "jax":
                params[name] = jnp.array(value)
            else:
                params[name] = value
        return params

    def _prepare(self, **kwargs):
        """Rewrite a python expression that consists of only basic math to use
        the prespecified math library. Santizes the string to prevent
        arbitrary code execution.

        Parameters
        ----------
        **kwargs : dict, optional
            Any additional keyword arguments, including the variable
            assignments.

        Returns
        -------
        tree : ast.*
            The root node of the parsed syntax tree.
        """
        tree = ast.parse(self.expression)

        # Walk the tree and confirm that it only contains the basic math.
        for node in ast.walk(tree):
            if isinstance(node, self._supported_ast_nodes):
                # Nothing to do, this is a valid operation for the ast.
                continue
            elif isinstance(node, ast.Call):
                # Check that function calls are only using items on the allow list.
                if node.func.id not in self._math_map:
                    raise ValueError(f"Unsupported function {node.func.id}")
            elif isinstance(node, ast.Name):
                if node.id in kwargs:
                    # This is a user supplied variable.
                    continue
                elif node.id in self._math_map:
                    # This is a math function or constant. Overwrite
                    if self.backend == "math":
                        node.id = self._math_map[node.id][0]
                    elif self.backend == "numpy":
                        node.id = self._math_map[node.id][1]
                    elif self.backend == "jax":
                        node.id = self._math_map[node.id][2]
                else:
                    raise ValueError(
                        f"Unrecognized named variable or function {node.id}. "
                        "This could be because the function is not supported or "
                        "you forgot to include the variable as an argument."
                    )
            else:
                raise ValueError(f"Invalid part of expression {type(node)}")

        # Convert the expression back into a string.
        self.expression = ast.unparse(tree)



from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest
from citation_compass import find_in_citations
from lightcurvelynx.astro_utils.pzflow_node import PZFlowNode
from lightcurvelynx.base_models import FunctionNode, ParameterizedNode
from lightcurvelynx.math_nodes.np_random import NumpyRandomFunc
from pzflow import Flow
from pzflow.bijectors import Reverse


class SumNode(ParameterizedNode):
    """A test class for the ParameterizedNode.

    Parameters
    ----------
    value1 : `float`
        The first value.
    value2 : `float`
        The second value.
    value_sum : `float`
        The sum of the two values.
    **kwargs : `dict`, optional
        Any additional keyword arguments.
    """

    def __init__(self, value1, value2, **kwargs):
        super().__init__(**kwargs)
        self.add_parameter("value1", value1, **kwargs)
        self.add_parameter("value2", value2, **kwargs)
        self.add_parameter(
            "value_sum",
            FunctionNode(self._sum, value1=self.value1, value2=self.value2),
            **kwargs,
        )

    def _sum(self, value1, value2):
        return value1 + value2


def test_pzflow_node_sample():
    """Test that we can sample numbers from a PZFlowNode."""
    flow = Flow(("x", "y"), Reverse())
    pz_node = PZFlowNode(flow, node_label="pznode")

    # Sample a bunch of parameters.
    state = pz_node.sample_parameters(num_samples=100)
    assert len(state["pznode"]) == 2
    assert len(state["pznode"]["x"]) == 100
    assert len(state["pznode"]["y"]) == 100

    # If we pass a seeded random number generator, we control the randomness.
    state1 = pz_node.sample_parameters(num_samples=50, rng_info=np.random.default_rng(seed=1))
    state2 = pz_node.sample_parameters(num_samples=50, rng_info=np.random.default_rng(seed=1))
    assert np.allclose(state1["pznode"]["x"], state2["pznode"]["x"])
    assert np.allclose(state1["pznode"]["y"], state2["pznode"]["y"])


def test_pzflow_node_chained():
    """Test that we can sample numbers from a PZFlowNode through another node."""
    flow = Flow(("a", "b"), Reverse())
    pz_node = PZFlowNode(flow, node_label="pznode")
    sum_node = SumNode(value1=pz_node.a, value2=pz_node.b, node_label="sum")

    # Sample a bunch of parameters.
    state = sum_node.sample_parameters(num_samples=25)
    assert len(state["sum"]["value_sum"]) == 25
    assert np.allclose(state["pznode"]["a"], state["sum"]["value1"])
    assert np.allclose(state["pznode"]["b"], state["sum"]["value2"])

    expected_sum = state["pznode"]["a"] + state["pznode"]["b"]
    assert np.allclose(state["sum"]["value_sum"], expected_sum)

    # Make sure that we only call the pzflow's sample() once during sampling of the node
    # (so all the outputs are consistent with each other). In other words, we do
    # not call compute once for each output variable.
    def _mock_sample(self, nsamples=1, conditions=None, save_conditions=True, seed=None):
        return pd.DataFrame({"a": [1], "b": [2]})

    with patch("pzflow.flow.Flow.sample", side_effect=_mock_sample) as mocked_sample:
        state = sum_node.sample_parameters(num_samples=1)
        mocked_sample.assert_called_once()


def test_pzflow_node_from_file(test_flow_filename):
    """Test that we can load and query a test flow."""
    pz_node = PZFlowNode.from_file(test_flow_filename, node_label="loaded_node")

    # Sample the pair of parameters defined by this flow (redshift and hostmass).
    state = pz_node.sample_parameters(num_samples=100)
    assert len(state["loaded_node"]) == 2
    assert len(state["loaded_node"]["redshift"]) == 100
    assert len(state["loaded_node"]["brightness"]) == 100


def test_conditional_pzflow_node(test_conditional_flow_filename):
    """Test that we can load and query a conditional flow."""
    redshift_node = NumpyRandomFunc(
        "uniform",
        low=0.05,
        high=0.3,
        node_label="redshift_node",
    )

    pz_node = PZFlowNode.from_file(
        test_conditional_flow_filename,
        node_label="loaded_node",
        redshift=redshift_node,
    )

    # Sample the pair of parameters defined by this flow (redshift and hostmass).
    state = pz_node.sample_parameters(num_samples=100)
    assert len(state["loaded_node"]) == 2
    assert len(state["loaded_node"]["redshift"]) == 100
    assert len(state["loaded_node"]["brightness"]) == 100


def test_invalid_conditional_pzflow_node(test_conditional_flow_filename):
    """Test that we raise an error if we try to create a pzflow node from
    a conditional flow, but do not provide the input parameters."""
    with pytest.raises(ValueError):
        # redshift is missing.
        _ = PZFlowNode.from_file(test_conditional_flow_filename, node_label="loaded_node")


def test_pzflow_node_citation(test_flow_filename):
    """Test that we can recover the citations for pzflow."""
    _ = PZFlowNode.from_file(test_flow_filename, node_label="loaded_node")
    citations = find_in_citations("PZFlowNode")
    for citation in citations:
        assert "Crenshaw et. al. 2024" in citation
